"""Metrics module.

This module contains classes that implement metrics used to monitor training sessions and evaluate models.
These metrics should all inherit from :class:`thelper.optim.metrics.Metric` to allow them to be dynamically
instantiated by the framework from a configuration file, and evaluated automatically inside a training
session. For more information on this, refer to :class:`thelper.train.base.Trainer`.
"""

import logging
from abc import abstractmethod
from typing import Any, Optional  # noqa: F401

import numpy as np
import sklearn.metrics
import torch

import thelper.concepts
import thelper.utils
from thelper.ifaces import ClassNamesHandler, PredictionConsumer

logger = logging.getLogger(__name__)


class Metric(PredictionConsumer):
    """Abstract metric interface.

    This interface defines basic functions required so that :class:`thelper.train.base.Trainer` can
    figure out how to instantiate, update, and optimize a given metric while training/evaluating a model.

    All metrics, by definition, must be 'optimizable'. This means that they should return a scalar value
    when 'evaluated' and define an optimal goal (-inf or +inf). If this is not possible, then the class
    should probably be derived using the more generic :class:`thelper.ifaces.PredictionConsumer`
    instead.
    """

    minimize = float("-inf")
    """Possible value of the ``goal`` attribute of this metric."""

    maximize = float("inf")
    """Possible value of the ``goal`` attribute of this metric."""

    @abstractmethod
    def update(self,        # see `thelper.typedefs.IterCallbackParams` for more info
               task,        # type: thelper.tasks.utils.Task
               input,       # type: thelper.typedefs.InputType
               pred,        # type: thelper.typedefs.AnyPredictionType
               target,      # type: thelper.typedefs.AnyTargetType
               sample,      # type: thelper.typedefs.SampleType
               loss,        # type: Optional[float]
               iter_idx,    # type: int
               max_iters,   # type: int
               epoch_idx,   # type: int
               max_epochs,  # type: int
               **kwargs,    # type: Any
               ):           # type: (...) -> None
        """Receives the latest prediction and groundtruth tensors from the training session.

        The data given here will be "consumed" internally, but it should NOT be modified. For example,
        a classification accuracy metric might accumulate the correct number of predictions in comparison
        to groundtruth labels, but never alter those predictions. The iteration/epoch indices may be
        used to 'reset' the internal state of this object when needed (for example, at the start of each
        new epoch).

        Remember that input, prediction, and target tensors received here will all have a batch dimension!

        The exact signature of this function should match the one of the callbacks defined in
        :class:`thelper.train.base.Trainer` and specified by ``thelper.typedefs.IterCallbackParams``.
        """
        raise NotImplementedError

    @abstractmethod
    def eval(self):
        """Returns the metric's evaluation result.

        The returned value should be a scalar. As a model improves, this scalar should get closer
        to the optimization goal (defined through the 'goal' attribute). This value will be queried
        at the end of each training epoch by the trainer.
        """
        raise NotImplementedError

    @property
    def goal(self):
        """Returns the scalar optimization goal of the metric.

        The returned goal can be the ``minimize`` or ``maximize`` members of ``thelper.optim.metrics.Metric``
        if the class's evaluation returns a scalar value, and ``None`` otherwise. The trainer will
        check this value to see if monitoring the metric's evaluation result progression is possible.
        """
        raise NotImplementedError

    @property
    def live_eval(self):
        """Returns whether this metric can/should be evaluated at every backprop iteration or not.

        By default, this returns ``True``, but implementations that are quite slow may return ``False``.
        """
        return True


@thelper.concepts.classification
@thelper.concepts.segmentation
class Accuracy(Metric):
    r"""Classification accuracy metric interface.

    This is a scalar metric used to monitor the label prediction accuracy of a model. By default,
    it works in ``top-k`` mode, meaning that the evaluation result is given by:

    .. math::
      \text{accuracy} = \frac{\text{nb. correct predictions}}{\text{nb. total predictions}} \cdot 100

    When :math:`k>1`, a 'correct' prediction is obtained if any of the model's top :math:`k` predictions
    (i.e. the :math:`k` predictions with the highest score) match the groundtruth label. Otherwise, if
    :math:`k=1`, then only the top prediction is compared to the groundtruth label. Note that for
    binary classification problems, :math:`k` should always be set to 1.

    This metric's goal is to maximize its value :math:`\in [0,100]` (a percentage is returned).

    Usage example inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example metric; it is used for lookup/printing only
            "top_5_accuracy": {
                # this type is used to instantiate the accuracy metric
                "type": "thelper.optim.metrics.Accuracy",
                # these parameters are passed to the wrapper's constructor
                "params": {
                    # the top prediction count to check for a match with the groundtruth
                    "top_k": 5
                }
            },
            # ...
        }
        # ...

    Todo: add support for 'dont care' target value?

    Attributes:
        top_k: number of top predictions to consider when matching with the groundtruth (default=1).
        max_win_size: maximum moving average window size to use (default=None, which equals dataset size).
        correct: total number of correct predictions stored using an array for window-based averaging.
        total: total number of predictions stored using an array for window-based averaging.
        warned_eval_bad: toggles whether the division-by-zero warning has been flagged or not.
    """

    def __init__(self, top_k=1, max_win_size=None):
        """Receives the number of predictions to consider for matches (``top_k``) and the moving average
        window size (``window_size``).

        Note that by default, if ``max_win_size`` is not provided here, the value given to ``max_iters`` on
        the first update call will be used instead to fix the sliding window length. In any case, the
        smallest of ``max_iters`` and ``max_win_size`` will be used to determine the actual window size.
        """
        assert isinstance(top_k, int) and top_k > 0, "invalid top-k value"
        assert max_win_size is None or (isinstance(max_win_size, int) and max_win_size > 0), \
            "invalid max sliding window size (should be positive integer)"
        self.top_k = top_k
        self.max_win_size = max_win_size
        self.correct = None  # will be instantiated on first iter
        self.total = None  # will be instantiated on first iter
        self.warned_eval_bad = False

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this metric."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(top_k={repr(self.top_k)}, max_win_size={repr(self.max_win_size)})"

    def update(self,        # see `thelper.typedefs.IterCallbackParams` for more info
               task,        # type: thelper.tasks.utils.Task
               input,       # type: thelper.typedefs.InputType
               pred,        # type: thelper.typedefs.ClassificationPredictionType
               target,      # type: thelper.typedefs.ClassificationTargetType
               sample,      # type: thelper.typedefs.SampleType
               loss,        # type: Optional[float]
               iter_idx,    # type: int
               max_iters,   # type: int
               epoch_idx,   # type: int
               max_epochs,  # type: int
               **kwargs,    # type: Any
               ):           # type: (...) -> None
        """Receives the latest class prediction and groundtruth labels from the training session.

        This function computes and accumulate the number of correct and total predictions in
        the internal arrays, cycling over the iteration index if the maximum window length is reached.

        The exact signature of this function should match the one of the callbacks defined in
        :class:`thelper.train.base.Trainer` and specified by ``thelper.typedefs.IterCallbackParams``.
        """
        assert len(kwargs) == 0, "unexpected extra arguments present in update call"
        assert iter_idx is not None and max_iters is not None and iter_idx < max_iters, \
            "bad iteration indices given to metric update function"
        curr_win_size = max_iters if self.max_win_size is None else min(self.max_win_size, max_iters)
        if self.correct is None or self.correct.size != curr_win_size:
            # each 'iteration' will have a corresponding bin with counts for that batch
            self.correct = np.zeros(curr_win_size, dtype=np.int64)
            self.total = np.zeros(curr_win_size, dtype=np.int64)
        curr_idx = iter_idx % curr_win_size
        if target is None or target.numel() == 0:
            # only accumulate results when groundtruth is available
            self.correct[curr_idx] = 0
            self.total[curr_idx] = 0
            return
        assert pred.dim() == target.dim() + 1, "prediction/gt tensors dim mismatch (should be BxCx[...] and Bx[...])"
        assert pred.shape[0] == target.shape[0], "prediction/gt tensors batch size mismatch"
        assert pred.dim() <= 2 or pred.shape[2:] == target.shape[1:], "prediction/gt tensors array size mismatch"
        top_k = pred.topk(self.top_k, dim=1)[1].view(pred.shape[0], self.top_k, -1).numpy()
        true_k = target.view(target.shape[0], 1, -1).expand(-1, self.top_k, -1).numpy()
        self.correct[curr_idx] = np.any(np.equal(top_k, true_k), axis=1).sum(dtype=np.int64)
        self.total[curr_idx] = target.numel()

    def eval(self):
        """Returns the current accuracy (in percentage) based on the accumulated prediction counts.

        Will issue a warning if no predictions have been accumulated yet.
        """
        if self.total is None or self.total.size == 0 or np.sum(self.total) == 0:
            if not self.warned_eval_bad:
                self.warned_eval_bad = True
                logger.warning("category accuracy eval result invalid (set as 0.0), no results accumulated")
            return 0.0
        return (float(np.sum(self.correct)) / float(np.sum(self.total))) * 100

    def reset(self):
        """Toggles a reset of the metric's internal state, deallocating count arrays."""
        self.correct = None
        self.total = None

    @property
    def goal(self):
        """Returns the scalar optimization goal of this metric (maximization)."""
        return Metric.maximize


@thelper.concepts.regression
class MeanAbsoluteError(Metric):
    r"""Mean absolute error metric interface.

    This is a scalar metric used to monitor the mean absolute deviation (or error) for a model's
    predictions. This regression metric can be described as:

    .. math::
        e(x, y) = E = \{e_1,\dots,e_N\}^\top, \quad
        e_n = \left| x_n - y_n \right|,

    where :math:`N` is the batch size. If ``reduction`` is not ``'none'``, then:

    .. math::
        \text{MAE}(x, y) =
        \begin{cases}
            \operatorname{mean}(E), & \text{if reduction } = \text{mean.}\\
            \operatorname{sum}(E),  & \text{if reduction } = \text{sum.}
        \end{cases}

    `x` and `y` are tensors of arbitrary shapes with a total of `n` elements each.

    Usage example inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example metric; it is used for lookup/printing only
            "mae": {
                # this type is used to instantiate the error metric
                "type": "thelper.optim.metrics.MeanAbsoluteError",
                "params": {
                    "reduction": "mean"
                }
            },
            # ...
        }
        # ...

    Todo: add support for 'dont care' target value?

    Attributes:
        max_win_size: maximum moving average window size to use (default=None, which equals dataset size).
        reduction: string representing the tensor reduction strategy to use.
        errors: array of error values stored for window-based averaging.
        warned_eval_bad: toggles whether the division-by-zero warning has been flagged or not.
    """

    def __init__(self, reduction="mean", max_win_size=None):
        """Receives the reduction strategy and the moving average window size (``window_size``).

        Note that by default, if ``max_win_size`` is not provided here, the value given to ``max_iters`` on
        the first update call will be used instead to fix the sliding window length. In any case, the
        smallest of ``max_iters`` and ``max_win_size`` will be used to determine the actual window size.
        """
        assert max_win_size is None or (isinstance(max_win_size, int) and max_win_size > 0), \
            "invalid max sliding window size (should be positive integer)"
        assert reduction != "none", "metric must absolutely return a scalar, must reduce"
        self.reduction = reduction
        self.max_win_size = max_win_size
        self.errors = None  # will be instantiated on first iter
        self.warned_eval_bad = False

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this metric."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(reduction={repr(self.reduction)}, max_win_size={repr(self.max_win_size)})"

    def update(self,        # see `thelper.typedefs.IterCallbackParams` for more info
               task,        # type: thelper.tasks.utils.Task
               input,       # type: thelper.typedefs.InputType
               pred,        # type: thelper.typedefs.ClassificationPredictionType
               target,      # type: thelper.typedefs.ClassificationTargetType
               sample,      # type: thelper.typedefs.SampleType
               loss,        # type: Optional[float]
               iter_idx,    # type: int
               max_iters,   # type: int
               epoch_idx,   # type: int
               max_epochs,  # type: int
               **kwargs,    # type: Any
               ):           # type: (...) -> None
        """Receives the latest predictions and target values from the training session.

        This function computes and accumulates the L1 distance between predictions and targets in the
        internal array, cycling over the iteration index if the maximum window length is reached.

        The exact signature of this function should match the one of the callbacks defined in
        :class:`thelper.train.base.Trainer` and specified by ``thelper.typedefs.IterCallbackParams``.
        """
        assert len(kwargs) == 0, "unexpected extra arguments present in update call"
        assert iter_idx is not None and max_iters is not None and iter_idx < max_iters, \
            "bad iteration indices given to metric update function"
        curr_win_size = max_iters if self.max_win_size is None else min(self.max_win_size, max_iters)
        if self.errors is None or self.errors.size != curr_win_size:
            # each 'iteration' will have a corresponding bin with the average L1 loss for that batch
            self.errors = np.asarray([None] * curr_win_size)
        curr_idx = iter_idx % curr_win_size
        if target is None or target.numel() == 0:
            # only accumulate results when groundtruth is available
            self.errors[curr_idx] = None
            return
        assert pred.shape == target.shape, "prediction/gt tensors shape mismatch"
        self.errors[curr_idx] = torch.nn.functional.l1_loss(pred, target, reduction=self.reduction).item()

    def eval(self):
        """Returns the current (average) mean absolute error based on the accumulated values.

        Will issue a warning if no predictions have been accumulated yet.
        """
        if self.errors is None or self.errors.size == 0 or len([d for d in self.errors if d is not None]) == 0:
            if not self.warned_eval_bad:
                self.warned_eval_bad = True
                logger.warning("mean absolute error eval result invalid (set as 0.0), no results accumulated")
            return 0.0
        return np.mean([d for d in self.errors if d is not None])

    def reset(self):
        """Toggles a reset of the metric's internal state, deallocating the errors array."""
        self.errors = None

    @property
    def goal(self):
        """Returns the scalar optimization goal of this metric (minimization)."""
        return Metric.minimize


@thelper.concepts.regression
class MeanSquaredError(Metric):
    r"""Mean squared error metric interface.

    This is a scalar metric used to monitor the mean squared deviation (or error) for a model's
    predictions. This regression metric can be described as:

    .. math::
        e(x, y) = E = \{e_1,\dots,e_N\}^\top, \quad
        e_n = \left( x_n - y_n \right)^2,

    where :math:`N` is the batch size. If ``reduction`` is not ``'none'``, then:

    .. math::
        \text{MSE}(x, y) =
        \begin{cases}
            \operatorname{mean}(E), & \text{if reduction } = \text{mean.}\\
            \operatorname{sum}(E),  & \text{if reduction } = \text{sum.}
        \end{cases}

    `x` and `y` are tensors of arbitrary shapes with a total of `n` elements each.

    Usage example inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example metric; it is used for lookup/printing only
            "mse": {
                # this type is used to instantiate the error metric
                "type": "thelper.optim.metrics.MeanSquaredError",
                "params": {
                    "reduction": "mean"
                }
            },
            # ...
        }
        # ...

    Todo: add support for 'dont care' target value?

    Attributes:
        max_win_size: maximum moving average window size to use (default=None, which equals dataset size).
        reduction: string representing the tensor reduction strategy to use.
        errors: array of error values stored for window-based averaging.
        warned_eval_bad: toggles whether the division-by-zero warning has been flagged or not.
    """

    def __init__(self, reduction="mean", max_win_size=None):
        """Receives the reduction strategy and the moving average window size (``window_size``).

        Note that by default, if ``max_win_size`` is not provided here, the value given to ``max_iters`` on
        the first update call will be used instead to fix the sliding window length. In any case, the
        smallest of ``max_iters`` and ``max_win_size`` will be used to determine the actual window size.
        """
        assert max_win_size is None or (isinstance(max_win_size, int) and max_win_size > 0), \
            "invalid max sliding window size (should be positive integer)"
        assert reduction != "none", "metric must absolutely return a scalar, must reduce"
        self.reduction = reduction
        self.max_win_size = max_win_size
        self.errors = None  # will be instantiated on first iter
        self.warned_eval_bad = False

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this metric."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(reduction={repr(self.reduction)}, max_win_size={repr(self.max_win_size)})"

    def update(self,        # see `thelper.typedefs.IterCallbackParams` for more info
               task,        # type: thelper.tasks.utils.Task
               input,       # type: thelper.typedefs.InputType
               pred,        # type: thelper.typedefs.ClassificationPredictionType
               target,      # type: thelper.typedefs.ClassificationTargetType
               sample,      # type: thelper.typedefs.SampleType
               loss,        # type: Optional[float]
               iter_idx,    # type: int
               max_iters,   # type: int
               epoch_idx,   # type: int
               max_epochs,  # type: int
               **kwargs,    # type: Any
               ):           # type: (...) -> None
        """Receives the latest predictions and target values from the training session.

        This function computes and accumulates the mean squared error between predictions and targets in
        the internal array, cycling over the iteration index if the maximum window length is reached.

        The exact signature of this function should match the one of the callbacks defined in
        :class:`thelper.train.base.Trainer` and specified by ``thelper.typedefs.IterCallbackParams``.
        """
        assert len(kwargs) == 0, "unexpected extra arguments present in update call"
        assert iter_idx is not None and max_iters is not None and iter_idx < max_iters, \
            "bad iteration indices given to metric update function"
        curr_win_size = max_iters if self.max_win_size is None else min(self.max_win_size, max_iters)
        if self.errors is None or self.errors.size != curr_win_size:
            # each 'iteration' will have a corresponding bin with the average MSE loss for that batch
            self.errors = np.asarray([None] * curr_win_size)
        curr_idx = iter_idx % curr_win_size
        if target is None or target.numel() == 0:
            # only accumulate results when groundtruth is available
            self.errors[curr_idx] = None
            return
        assert pred.shape == target.shape, "prediction/gt tensors shape mismatch"
        self.errors[curr_idx] = torch.nn.functional.mse_loss(pred, target, reduction=self.reduction).item()

    def eval(self):
        """Returns the current (average) mean squared error based on the accumulated values.

        Will issue a warning if no predictions have been accumulated yet.
        """
        if self.errors is None or self.errors.size == 0 or len([d for d in self.errors if d is not None]) == 0:
            if not self.warned_eval_bad:
                self.warned_eval_bad = True
                logger.warning("mean squared error eval result invalid (set as 0.0), no results accumulated")
            return 0.0
        return np.mean([d for d in self.errors if d is not None])

    def reset(self):
        """Toggles a reset of the metric's internal state, deallocating the errors array."""
        self.errors = None

    @property
    def goal(self):
        """Returns the scalar optimization goal of this metric (minimization)."""
        return Metric.minimize


@thelper.concepts.classification
@thelper.concepts.segmentation
class ExternalMetric(Metric, ClassNamesHandler):
    r"""External metric wrapping interface.

    This interface is used to wrap external metrics and use them in the training framework. The metrics
    of ``sklearn.metrics`` are good candidates that have been used extensively with this interface in
    the past, but those of other libraries might also be compatible.

    Along with the name of the class to import and its constructor's parameters, the user must provide
    a handling mode that specifies how prediction and groundtruth data should be handled in this wrapper.
    Also, extra arguments such as target label names, goal information, and window sizes can be provided
    for specific use cases related to the selected handling mode.

    For now, two metric handling modes (both related to classification) are supported:

      * ``classif_best``: the wrapper will accumulate the predicted and groundtruth classification \
        labels forwarded by the trainer and provide them to the external metric for evaluation. If \
        a target label name is specified, then only classifications related to that label will be \
        accumulated. This is the handling mode required for count-based classification metrics such \
        as accuracy, F-Measure, precision, recall, etc.

      * ``classif_score``: the wrapper will accumulate the prediction score of the targeted label \
        along with a boolean that indicates whether this label was the groundtruth label or not. This \
        is the handling mode required for score-based classification metrics such as when computing \
        the area under the ROC curve (AUC).

    Usage examples inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the first example metric; it is used for lookup/printing only
            "f1_score_reject": {
                # this type is used to instantiate the wrapper
                "type": "thelper.optim.metrics.ExternalMetric",
                # these parameters are passed to the wrapper's constructor
                "params": {
                    # the external class to import
                    "metric_name": "sklearn.metrics.f1_score",
                    # the parameters passed to the external class's constructor
                    "metric_params": {},
                    # the wrapper metric handling mode
                    "metric_type": "classif_best",
                    # the target class name (note: dataset-specific)
                    "target_name": "reject",
                    # the goal type of the external metric
                    "metric_goal": "max"
                }
            },
            # this is the name of the second example metric; it is used for lookup/printing only
            "roc_auc_accept": {
                # this type is used to instantiate the wrapper
                "type": "thelper.optim.metrics.ExternalMetric",
                # these parameters are passed to the wrapper's constructor
                "params": {
                    # the external class to import
                    "metric_name": "sklearn.metrics.roc_auc_score",
                    # the parameters passed to the external class's constructor
                    "metric_params": {},
                    # the wrapper metric handling mode
                    "metric_type": "classif_score",
                    # the target class name (note: dataset-specific)
                    "target_name": "accept",
                    # the goal type of the external metric
                    "metric_goal": "max"
                }
            },
            # ...
        }
        # ...

    Attributes:
        metric_goal: goal of the external metric, used for monitoring. Can be ``min`` or ``max``.
        metric_type: handling mode of the external metric. Can only be one of the predetermined values.
        metric: type of the external metric that will be instantiated when ``eval`` is called.
        metric_params: dictionary of parameters passed to the external metric on instantiation.
        target_name: name of the targeted label. Used only in handling modes related to classification.
        target_idx: index of the targeted label. Used only in handling modes related to classification.
        class_names: holds the list of class label names provided by the dataset parser. If it is not
            provided when the constructor is called, it will be set by the trainer at runtime.
        force_softmax: specifies whether a softmax operation should be applied to the prediction scores
            obtained from the trainer. Only used with the "classif_score" handling mode.
        max_win_size: maximum moving average window size to use (default=None, which equals dataset size).
        pred: queue used to store predictions-related values for window-based averaging.
        target: queue used to store groundtruth-related values for window-based averaging.
    """

    def __init__(self, metric_name, metric_type, metric_goal, metric_params=None, target_name=None,
                 class_names=None, max_win_size=None, force_softmax=True, live_eval=True):
        """Receives all necessary arguments for wrapper initialization and external metric instantiation.

        See :class:`thelper.optim.metrics.ExternalMetric` for information on arguments.
        """
        assert isinstance(metric_name, str), "metric_name must be fully qualifiied class name to import"
        assert metric_params is None or isinstance(metric_params, dict), "metric_params must be dictionary"
        supported_handling_types = [
            "classif_top1", "classif_best",  # the former is for backwards-compat with the latter
            "classif_scores", "classif_score",  # the former is for backwards-compat with the latter
            "regression",  # missing impl, work in progress @@@ TODO
        ]
        assert isinstance(metric_type, str) and metric_type in supported_handling_types, \
            f"unknown metric type {repr(metric_type)}"
        if metric_type == "classif_top1":
            metric_type = "classif_best"  # they are identical, just overwrite for backwards compat
        if metric_type == "classif_scores":
            metric_type = "classif_score"  # they are identical, just overwrite for backwards compat
        assert metric_goal is not None and metric_goal in ["max", "min"], "unexpected goal type"
        self.metric_goal = Metric.maximize if metric_goal == "max" else Metric.minimize
        self.metric_type = metric_type
        self.metric_name = metric_name
        self.metric = thelper.utils.import_class(metric_name)
        self.metric_params = metric_params if metric_params is not None else {}
        self.target_name = target_name
        self.target_idx = None
        self.force_softmax = None
        if metric_type == "classif_score":
            self.force_softmax = force_softmax  # only useful in this case
        # elif "regression" in metric_type: missing impl for custom handling @@@
        assert max_win_size is None or (isinstance(max_win_size, int) and max_win_size > 0), \
            "invalid max sliding window size (should be positive integer)"
        self.max_win_size = max_win_size
        self.pred = None  # will be instantiated on first iter
        self.target = None  # will be instantiated on first iter
        self._live_eval = live_eval  # could be 'False' for external impls that are pretty slow to eval
        ClassNamesHandler.__init__(self, class_names)

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this metric."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(metric_name={repr(self.metric_name)}, metric_type={repr(self.metric_type)}, " + \
            f"metric_goal={'min' if self.goal == Metric.minimize else 'max'}, " + \
            f"metric_params={repr(self.metric_params)}, target_name={repr(self.target_name)}, " + \
            f"class_names={repr(self.class_names)}, max_win_size={repr(self.max_win_size)}, " + \
            f"force_softmax={repr(self.force_softmax)})"

    @ClassNamesHandler.class_names.setter
    def class_names(self, class_names):
        """Sets the class label names that must be predicted by the model.

        This is only useful in metric handling modes related to classification. The goal of having
        class names here is to translate a target class label (provided in the constructor) into a
        target class index. This is required as predictions are not mapped to their original names
        (in string format) before being forwarded to this object by the trainer.
        """
        if "classif" in self.metric_type:
            ClassNamesHandler.class_names.fset(self, class_names)
            if self.target_name is not None and self.class_names is not None:
                assert self.target_name in self.class_indices, \
                    f"could not find target name {repr(self.target_name)} in class names list"
                self.target_idx = self.class_indices[self.target_name]
            else:
                self.target_idx = None

    def update(self,        # see `thelper.typedefs.IterCallbackParams` for more info
               task,        # type: thelper.tasks.utils.Task
               input,       # type: thelper.typedefs.InputType
               pred,        # type: thelper.typedefs.ClassificationTargetType
               target,      # type: thelper.typedefs.ClassificationPredictionType
               sample,      # type: thelper.typedefs.SampleType
               loss,        # type: Optional[float]
               iter_idx,    # type: int
               max_iters,   # type: int
               epoch_idx,   # type: int
               max_epochs,  # type: int
               **kwargs,    # type: Any
               ):           # type: (...) -> None
        """Receives the latest predictions and target values from the training session.

        The handling of the data received here will depend on the current metric's handling mode.

        The exact signature of this function should match the one of the callbacks defined in
        :class:`thelper.train.base.Trainer` and specified by ``thelper.typedefs.IterCallbackParams``.
        """
        assert len(kwargs) == 0, "unexpected extra arguments present in update call"
        assert iter_idx is not None and max_iters is not None and iter_idx < max_iters, \
            "bad iteration indices given to metric update function"
        curr_win_size = max_iters if self.max_win_size is None else min(self.max_win_size, max_iters)
        if self.pred is None or self.pred.size != curr_win_size:
            # each 'iteration' will have a corresponding bin with counts for that batch
            self.pred = np.asarray([None] * curr_win_size)
            self.target = np.asarray([None] * curr_win_size)
        curr_idx = iter_idx % curr_win_size
        if "classif" in self.metric_type:
            if hasattr(task, "class_names") and task.class_names != self.class_names:
                self.class_names = task.class_names
            if target is None or target.numel() == 0:
                # only accumulate results when groundtruth is available
                self.pred[curr_idx] = None
                self.target[curr_idx] = None
                return
            assert self.target_name is None or self.target_idx is not None, \
                f"could not map target name '{self.target_name}' to target idx, missing class list"
            assert pred.dim() == 2 or target.dim() == 1, "current ext metric implementation only supports batched 1D outputs"
            assert pred.shape[0] == target.shape[0], "prediction/gt tensors batch size mismatch"
            if self.target_idx is not None:
                pred_label = pred.topk(1, dim=1)[1].view(pred.shape[0])
                y_true, y_pred = [], []
                if self.metric_type == "classif_best":
                    assert pred_label.numel() == target.numel(), "pred/target classification element count mismatch"
                    must_keep = [y_pred == self.target_idx or y_true == self.target_idx for y_pred, y_true in zip(pred_label, target)]
                    for idx, keep in enumerate(must_keep):
                        if keep:
                            y_true.append(target[idx].item() == self.target_idx)
                            y_pred.append(pred_label[idx].item() == self.target_idx)
                else:  # self.metric_type == "classif_score"
                    if self.force_softmax:
                        with torch.no_grad():
                            pred = torch.nn.functional.softmax(pred, dim=1)
                    for idx, tgt in enumerate(target):
                        y_true.append(tgt.item() == self.target_idx)
                        y_pred.append(pred[idx, self.target_idx].item())
                self.target[curr_idx] = y_true
                self.pred[curr_idx] = y_pred
            else:
                assert self.metric_type != "classif_score", "score-based classif analysis (e.g. roc auc) must specify target label"
                if self.metric_type == "classif_best":
                    self.target[curr_idx] = [target[idx].item() for idx in range(pred.numel())]
                    self.pred[curr_idx] = [pred[idx].item() for idx in range(pred.numel())]
        else:  # if self.metric_type == "regression":
            raise NotImplementedError

    def eval(self):
        """Returns the external metric's evaluation result."""
        if "classif" in self.metric_type:
            assert self.target.size == self.pred.size, "internal window size mismatch"
            pred, target = zip(*[(pred, target) for preds, targets in zip(self.pred, self.target)
                                 if targets is not None for pred, target in zip(preds, targets)])
            return self.metric(np.stack(target, axis=0), np.stack(pred, axis=0), **self.metric_params)
        else:  # if self.metric_type == "regression":
            raise NotImplementedError

    def reset(self):
        """Toggles a reset of the metric's internal state, emptying pred/target queues."""
        self.pred = None
        self.target = None

    @property
    def goal(self):
        """Returns the scalar optimization goal of this metric (user-defined)."""
        return self.metric_goal

    @property
    def live_eval(self):
        """Returns whether this metric can/should be evaluated at every backprop iteration or not.

        By default, this returns ``True``, but implementations that are quite slow may return ``False``.
        """
        return self._live_eval


@thelper.concepts.classification
@thelper.concepts.segmentation
class ROCCurve(Metric, ClassNamesHandler):
    """Receiver operating characteristic (ROC) computation interface.

    This class provides an interface to ``sklearn.metrics.roc_curve`` and ``sklearn.metrics.roc_auc_score``
    that can produce various types of ROC-related information including the area under the curve (AUC), the
    false positive and negative rates for various operating points, and the ROC curve itself as an image
    (also compatible with tensorboardX).

    By default, evaluating this metric returns the Area Under the Curve (AUC). If a target operating point is
    set, it will instead return the false positive/negative prediction rate of the model at that point.

    Usage examples inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the first example; it will output the AUC of the "reject" class
            "roc_reject_auc": {
                # this type is used to instantiate the ROC metric
                "type": "thelper.optim.metrics.ROCCurve",
                # these parameters are passed to the constructor
                "params": {
                    # the name of the class to evaluate
                    "target_name": "reject"
                }
            },
            # this is the name of the second example; it will output the FPR at TPR=0.99
            "roc_reject_0.99tpr": {
                # this type is used to instantiate the ROC metric
                "type": "thelper.optim.metrics.ROCCurve",
                # these parameters are passed to the constructor
                "params": {
                    # the name of the class to evaluate
                    "target_name": "reject",
                    # the target true positive rate (TPR) operating point
                    "target_tpr": 0.99
                }
            },
            # ...
        }
        # ...

    Attributes:
        target_inv: used to target all classes except the named one(s); experimental!
        target_name: name of targeted class to generate the roc curve/auc information for.
        target_tpr: target operating point in terms of true positive rate (provided in constructor).
        target_fpr: target operating point in terms of false positive rate (provided in constructor).
        target_idx: index of the targeted class, mapped from target_name using the class_names list.
        class_names: holds the list of class label names provided by the dataset parser. If it is not
            provided when the constructor is called, it will be set by the trainer at runtime.
        force_softmax: specifies whether a softmax operation should be applied to the prediction scores
            obtained from the trainer.
        curve: roc curve generator function, called at evaluation time to generate the output string.
        auc: auc score generator function, called at evaluation time to generate the output string.
        score: queue used to store prediction score values for window-based averaging.
        true: queue used to store groundtruth label values for window-based averaging.
    """

    def __init__(self, target_name, target_tpr=None, target_fpr=None, class_names=None,
                 force_softmax=True, sample_weight=None, drop_intermediate=True):
        """Receives the target class/operating point info, log parameters, and roc computation arguments.

        Args:
            target_name: name of targeted class to generate the roc curve/auc information for.
            target_tpr: target operating point in terms of true positive rate (provided in constructor).
            target_fpr: target operating point in terms of false positive rate (provided in constructor).
            class_names: holds the list of class label names provided by the dataset parser. If it is not
                provided when the constructor is called, it will be set by the trainer at runtime.
            force_softmax: specifies whether a softmax operation should be applied to the prediction scores
                obtained from the trainer.
            sample_weight: passed to ``sklearn.metrics.roc_curve`` and ``sklearn.metrics.roc_auc_score``.
            drop_intermediate: passed to ``sklearn.metrics.roc_curve``.
        """
        assert target_name is not None, "must provide a target (class) name for ROC metric"
        self.target_inv = False
        if isinstance(target_name, str) and target_name[0] == "!":
            self.target_inv = True
            self.target_name = target_name.split("!", 1)[1]
        else:
            self.target_name = target_name
        self.target_tpr, self.target_fpr = None, None
        assert target_tpr is None or target_fpr is None, "must specify only one of target_fpr and target_tpr, not both"
        if target_tpr is not None or target_fpr is not None:
            target_xpr = target_tpr if target_tpr is not None else target_fpr
            assert isinstance(target_xpr, float), "expected float type for target operating point"
            assert 0 <= target_xpr <= 1, "invalid target operation point value (must be in [0,1])"
            if target_tpr is not None:
                self.target_tpr = target_tpr
            else:  # if target_fpr is not None
                self.target_fpr = target_fpr
        self.target_idx = None
        self.force_softmax = force_softmax
        self.sample_weight = sample_weight
        self.drop_intermediate = drop_intermediate

        def gen_curve(y_true, y_score, _target_idx, _target_inv, _sample_weight=sample_weight, _drop_intermediate=drop_intermediate):
            assert _target_idx is not None, "missing positive target idx at run time"
            _y_true, _y_score = [], []
            for sample_idx, label_idx in enumerate(y_true):
                _y_true.append(label_idx != _target_idx if _target_inv else label_idx == _target_idx)
                _y_score.append(1 - y_score[sample_idx, _target_idx] if _target_inv else y_score[sample_idx, _target_idx])
            res = sklearn.metrics.roc_curve(_y_true, _y_score, sample_weight=_sample_weight, drop_intermediate=_drop_intermediate)
            return res

        def gen_auc(y_true, y_score, _target_idx, _target_inv, _sample_weight=sample_weight):
            assert _target_idx is not None, "missing positive target idx at run time"
            _y_true, _y_score = [], []
            for sample_idx, label_idx in enumerate(y_true):
                _y_true.append(label_idx != _target_idx if _target_inv else label_idx == _target_idx)
                _y_score.append(1 - y_score[sample_idx, _target_idx] if _target_inv else y_score[sample_idx, _target_idx])
            res = sklearn.metrics.roc_auc_score(_y_true, _y_score, sample_weight=_sample_weight)
            return res

        self.curve = gen_curve
        self.auc = gen_auc
        self.score = None
        self.true = None
        ClassNamesHandler.__init__(self, class_names)

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this metric."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(target_name={repr(self.target_name)}, target_tpr={repr(self.target_tpr)}, " + \
            f"target_fpr={repr(self.target_fpr)}, class_names={repr(self.class_names)}, " + \
            f"force_softmax={repr(self.force_softmax)}, sample_weight={repr(self.sample_weight)}, " + \
            f"drop_intermediate={repr(self.drop_intermediate)})"

    @ClassNamesHandler.class_names.setter
    def class_names(self, class_names):
        """Sets the class label names that must be predicted by the model."""
        ClassNamesHandler.class_names.fset(self, class_names)
        if self.target_name is not None and self.class_names is not None:
            assert self.target_name in self.class_indices, \
                f"could not find target name {repr(self.target_name)} in class names list"
            self.target_idx = self.class_indices[self.target_name]
        else:
            self.target_idx = None

    def update(self,        # see `thelper.typedefs.IterCallbackParams` for more info
               task,        # type: thelper.tasks.utils.Task
               input,       # type: thelper.typedefs.InputType
               pred,        # type: thelper.typedefs.ClassificationPredictionType
               target,      # type: thelper.typedefs.ClassificationTargetType
               sample,      # type: thelper.typedefs.SampleType
               loss,        # type: Optional[float]
               iter_idx,    # type: int
               max_iters,   # type: int
               epoch_idx,   # type: int
               max_epochs,  # type: int
               **kwargs,    # type: Any
               ):           # type: (...) -> None
        """Receives the latest predictions and target values from the training session.

        The exact signature of this function should match the one of the callbacks defined in
        :class:`thelper.train.base.Trainer` and specified by ``thelper.typedefs.IterCallbackParams``.
        """
        assert len(kwargs) == 0, "unexpected extra arguments present in update call"
        assert isinstance(task, thelper.tasks.Classification), "roc curve only impl for classif tasks"
        assert iter_idx is not None and max_iters is not None and iter_idx < max_iters, \
            "bad iteration indices given to metric update function"
        if self.score is None or self.score.size != max_iters:
            self.score = np.asarray([None] * max_iters)
            self.true = np.asarray([None] * max_iters)
        if task.class_names != self.class_names:
            self.class_names = task.class_names
        if target is None or target.numel() == 0:
            # only accumulate results when groundtruth is available
            self.score[iter_idx] = None
            self.true[iter_idx] = None
            return
        assert pred.dim() == 2 or target.dim() == 1, "current classif report impl only supports batched 1D outputs"
        assert pred.shape[0] == target.shape[0], "prediction/gt tensors batch size mismatch"
        assert pred.shape[1] == len(self.class_names), "unexpected prediction class dimension size"
        if self.force_softmax:
            with torch.no_grad():
                pred = torch.nn.functional.softmax(pred, dim=1)
        self.score[iter_idx] = pred.numpy()
        self.true[iter_idx] = target.numpy()

    def eval(self):
        """Returns the evaluation result (AUC/TPR/FPR).

        If no target operating point is set, the returned value is the AUC for the target class. If a
        target TPR is set, the returned value is the FPR for that operating point. If a target FPR is set,
        the returned value is the TPR for that operating point.
        """
        if self.score is None or self.true is None:
            return None
        score, true = zip(*[(score, true) for scores, trues in zip(self.score, self.true)
                            if trues is not None for score, true in zip(scores, trues)])
        # if we did not specify a target operating point in terms of true/false positive rate, return AUC
        if self.target_tpr is None and self.target_fpr is None:
            return self.auc(np.stack(true, axis=0), np.stack(score, axis=0), self.target_idx, self.target_inv)
        # otherwise, find the opposite rate at the requested target operating point
        _fpr, _tpr, _thrs = self.curve(np.stack(true, axis=0), np.stack(score, axis=0), self.target_idx,
                                       self.target_inv, _drop_intermediate=False)
        for fpr, tpr, thrs in zip(_fpr, _tpr, _thrs):
            if self.target_tpr is not None and tpr >= self.target_tpr:
                # print("for target tpr = %.5f, fpr = %.5f at threshold = %f" % (self.target_tpr, fpr, thrs))
                return fpr
            elif self.target_fpr is not None and fpr >= self.target_fpr:
                # print("for target fpr = %.5f, tpr = %.5f at threshold = %f" % (self.target_fpr, tpr, thrs))
                return tpr
        # if we did not find a proper rate match above, return worse possible value
        if self.target_tpr is not None:
            # print("for target tpr = %.5f, fpr = 1.0 at threshold = min" % self.target_tpr)
            return 1.0
        else:  # if self.target_fpr is not None:
            # print("for target fpr = %.5f, tpr = 0.0 at threshold = max" % self.target_fpr)
            return 0.0

    def render(self):
        """Returns the ROC curve as a numpy-compatible RGBA image drawn by pyplot."""
        if self.score is None:
            return None
        score, true = zip(*[(score, true) for scores, trues in zip(self.score, self.true)
                            if trues is not None for score, true in zip(scores, trues)])
        fpr, tpr, t = self.curve(np.stack(true, axis=0), np.stack(score, axis=0), self.target_idx, self.target_inv)
        try:
            fig, ax = thelper.draw.draw_roc_curve(fpr, tpr)
            array = thelper.draw.fig2array(fig)
            return array
        except AttributeError as e:
            logger.warning(f"failed to render roc curve; caught exception:\n{str(e)}")
            # return None if rendering fails (probably due to matplotlib on displayless server)
            return None

    def reset(self):
        """Toggles a reset of the metric's internal state, emptying queues."""
        self.score = None
        self.true = None

    @property
    def goal(self):
        """Returns the scalar optimization goal of this metric (variable based on target op point)."""
        # if we did not specify a target operating point in terms of true/false positive rate, return AUC
        if self.target_tpr is None and self.target_fpr is None:
            return Metric.maximize  # AUC must be maximized
        if self.target_tpr is not None:
            return Metric.minimize  # fpr must be minimized
        else:  # if self.target_fpr is not None:
            return Metric.maximize  # tpr must be maximized

    @property
    def live_eval(self):
        """Returns whether this metric can/should be evaluated at every backprop iteration or not."""
        return False  # some operating modes might be pretty slow, check back impl later


@thelper.concepts.regression
class PSNR(Metric):
    r"""Peak Signal-to-Noise Ratio (PSNR) metric interface.

    This is a scalar metric used to monitor the change in quality of a signal (or image) following a
    transformation. For more information, see its definition on `[Wikipedia]`__.

    .. __: https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio

    The PSNR (in decibels, dB) between a modified signal :math:`x` and its original version :math:`y` is
    defined as:

    .. math::
        \text{PSNR}(x, y) = 10 * \log_{10} \Bigg( \frac{R^2}{\text{MSE}(x, y)} \Bigg)

    where :math:`\text{MSE}(x, y)` returns the mean squared error (see :class:`thelper.optim.metrics.MeanSquaredError`
    for more information), and :math:`R` is the maximum possible value for a single element in the input signal
    (i.e. its maximum "range").

    Usage example inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example metric; it is used for lookup/printing only
            "psnr": {
                # this type is used to instantiate the metric
                "type": "thelper.optim.metrics.PSNR",
                "params": {
                    "data_range": "255"
                }
            },
            # ...
        }
        # ...

    Attributes:
        max_win_size: maximum moving average window size to use (default=None, which equals dataset size).
        data_range: maximum value of an element in the target signal.
        psnrs: array of psnr values stored for window-based averaging.
        warned_eval_bad: toggles whether the division-by-zero warning has been flagged or not.
    """

    def __init__(self, data_range=1.0, max_win_size=None):
        """Receives all necessary initialization arguments to compute signal PSNRs,

        See :class:`thelper.optim.metrics.PSNR` for information on arguments.
        """
        self.max_win_size = max_win_size
        self.psnrs = None  # will be instantiated on first iter
        self.warned_eval_bad = False
        self.data_range = data_range

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this metric."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(data_range={repr(self.data_range)}, max_win_size={repr(self.max_win_size)})"

    def update(self,        # see `thelper.typedefs.IterCallbackParams` for more info
               task,        # type: thelper.tasks.utils.Task
               input,       # type: thelper.typedefs.InputType
               pred,        # type: thelper.typedefs.ClassificationPredictionType
               target,      # type: thelper.typedefs.ClassificationTargetType
               sample,      # type: thelper.typedefs.SampleType
               loss,        # type: Optional[float]
               iter_idx,    # type: int
               max_iters,   # type: int
               epoch_idx,   # type: int
               max_epochs,  # type: int
               **kwargs,    # type: Any
               ):           # type: (...) -> None
        """Receives the latest predictions and target values from the training session.

        The exact signature of this function should match the one of the callbacks defined in
        :class:`thelper.train.base.Trainer` and specified by ``thelper.typedefs.IterCallbackParams``.
        """
        assert len(kwargs) == 0, "unexpected extra arguments present in update call"
        assert iter_idx is not None and max_iters is not None and iter_idx < max_iters, \
            "bad iteration indices given to metric update function"
        curr_win_size = max_iters if self.max_win_size is None else min(self.max_win_size, max_iters)
        if self.psnrs is None or self.psnrs.size != curr_win_size:
            # each 'iteration' will have a corresponding bin with the psnr for that batch
            self.psnrs = np.asarray([None] * curr_win_size)
        curr_idx = iter_idx % curr_win_size
        if target is None or target.numel() == 0:
            # only accumulate results when groundtruth is available
            self.psnrs[curr_idx] = None
            return
        assert pred.shape == target.shape, "prediction/gt tensors shape mismatch"
        mse = np.mean(np.square(pred.numpy() - target.numpy()), dtype=np.float64)
        self.psnrs[curr_idx] = 10 * np.log10(self.data_range / mse)

    def eval(self):
        """Returns the current (average) PSNR based on the accumulated values.

        Will issue a warning if no predictions have been accumulated yet.
        """
        if self.psnrs is None or self.psnrs.size == 0 or len([v for v in self.psnrs if v is not None]) == 0:
            if not self.warned_eval_bad:
                self.warned_eval_bad = True
                logger.warning("psnr eval result invalid (set as 0.0), no results accumulated")
            return 0.0
        return np.mean([v for v in self.psnrs if v is not None])

    def reset(self):
        """Toggles a reset of the metric's internal state, deallocating the psnrs array."""
        self.psnrs = None

    @property
    def goal(self):
        """Returns the scalar optimization goal of this metric (maximization)."""
        return Metric.maximize


@thelper.concepts.detection
class AveragePrecision(Metric):
    r"""Object detection average precision score from PascalVOC.

    This metric is computed based on the evaluator function implemented in :mod:`thelper.optim.eval`.
    It can target a single class at a time, or produce the mean average precision for all classes.

    Usage example inside a session configuration file::

        # ...
        # lists all metrics to instantiate as a dictionary
        "metrics": {
            # ...
            # this is the name of the example metric; it is used for lookup/printing only
            "mAP": {
                # this type is used to instantiate the AP metric
                "type": "thelper.optim.metrics.AveragePrecision",
                # these parameters are passed to the wrapper's constructor
                "params": {
                    # no parameters means we will compute the mAP
                }
            },
            # ...
        }
        # ...

    Attributes:
        target_class: name of the class to target; if 'None', will compute mAP instead of AP.
        iou_threshold: Intersection Over Union (IOU) threshold for true/false positive classification.
        method: the evaluation method to use; can be the the latest & official PASCAL VOC toolkit
            approach ("all-points"), or the 11-point approach ("11-points") described in the original
            paper ("The PASCAL Visual Object Classes(VOC) Challenge").
        max_win_size: maximum moving average window size to use (default=None, which equals dataset size).
        preds: array holding the predicted bounding boxes for all input samples.
        targets: array holding the target bounding boxes for all input samples.
    """

    def __init__(self, target_class=None, iou_threshold=0.5, method="all-points", max_win_size=None):
        """Initializes metric attributes.

        Note that by default, if ``max_win_size`` is not provided here, the value given to ``max_iters`` on
        the first update call will be used instead to fix the sliding window length. In any case, the
        smallest of ``max_iters`` and ``max_win_size`` will be used to determine the actual window size.
        """
        assert max_win_size is None or (isinstance(max_win_size, int) and max_win_size > 0), \
            "invalid max sliding window size (should be positive integer)"
        self.target_class = target_class
        self.iou_threshold = iou_threshold
        self.method = method
        self.max_win_size = max_win_size
        self.preds = None  # will be instantiated on first iter
        self.targets = None  # will be instantiated on first iter
        self.task = None

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this metric."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(target_class={repr(self.target_class)}, max_win_size={repr(self.max_win_size)})"

    def update(self,        # see `thelper.typedefs.IterCallbackParams` for more info
               task,        # type: thelper.tasks.utils.Task
               input,       # type: thelper.typedefs.InputType
               pred,        # type: thelper.typedefs.DetectionPredictionType
               target,      # type: thelper.typedefs.DetectionTargetType
               sample,      # type: thelper.typedefs.SampleType
               loss,        # type: Optional[float]
               iter_idx,    # type: int
               max_iters,   # type: int
               epoch_idx,   # type: int
               max_epochs,  # type: int
               **kwargs):   # type: (...) -> None
        """Receives the latest bbox predictions and targets from the training session.

        The exact signature of this function should match the one of the callbacks defined in
        :class:`thelper.train.base.Trainer` and specified by ``thelper.typedefs.IterCallbackParams``.
        """
        assert len(kwargs) == 0, "unexpected extra arguments present in update call"
        assert iter_idx is not None and max_iters is not None and iter_idx < max_iters, \
            "bad iteration indices given to metric update function"
        curr_win_size = max_iters if self.max_win_size is None else min(self.max_win_size, max_iters)
        if self.preds is None or self.preds.size != curr_win_size:
            # each 'iteration' will have a corresponding bin with counts for that batch
            self.preds = np.asarray([None] * curr_win_size)
            self.targets = np.asarray([None] * curr_win_size)
        curr_idx = iter_idx % curr_win_size
        self.task = task  # keep reference for eval only
        if target is None or len(target) == 0:
            # only accumulate results when groundtruth is available (should we though? affects false negative count)
            self.preds[curr_idx] = None
            self.targets[curr_idx] = None
            return
        if not pred:
            pred = [[]] * len(target)
        assert isinstance(pred, list) and isinstance(target, list)
        assert all([isinstance(b, list) and
                    all([isinstance(p, thelper.tasks.detect.BoundingBox) for p in b]) for b in pred])
        assert all([isinstance(b, list) and
                    all([isinstance(t, thelper.tasks.detect.BoundingBox) for t in b]) for b in target])
        self.preds[curr_idx] = pred
        self.targets[curr_idx] = target

    def eval(self):
        """Returns the current accuracy (in percentage) based on the accumulated prediction counts.

        Will issue a warning if no predictions have been accumulated yet.
        """
        assert self.targets.size == self.preds.size, "internal window size mismatch"
        pred, target = zip(*[(pred, target) for preds, targets in zip(self.preds, self.targets)
                             if targets for pred, target in zip(preds, targets)])
        # maybe need to concat?
        pred, target = np.concatenate(pred), np.concatenate(target)  # possible due to image ids
        if len(pred) == 0:  # no predictions made by model
            return float("nan")
        metrics = thelper.optim.eval.compute_pascalvoc_metrics(pred, target, self.task,
                                                               self.iou_threshold, self.method)
        if self.target_class is None:
            # compute mAP wrt classes that have at least one positive sample
            return np.mean([m["AP"] for m in metrics.values() if m["total positives"] > 0])
        return metrics[self.target_class]["AP"]

    def reset(self):
        """Toggles a reset of the metric's internal state, deallocating bbox arrays."""
        self.preds = None
        self.targets = None

    @property
    def goal(self):
        """Returns the scalar optimization goal of this metric (maximization)."""
        return Metric.maximize

    @property
    def live_eval(self):
        """Returns whether this metric can/should be evaluated at every backprop iteration or not."""
        return False  # the current PascalVOC implementation is preeetty slow with lots of bboxes



"""Transformation composers module.

All transforms should aim to be compatible with both numpy arrays and
PyTorch tensors. By default, images are processed using ``__call__``,
meaning that for a given transformation ``t``, we apply it via::

    image_transformed = t(image)

All important parameters for an operation should also be passed in the
constructor and exposed in the operation's ``__repr__`` function so that
external parsers can discover exactly how to reproduce their behavior. For
now, these representations are used for debugging more than anything else.
"""

import bisect
import logging

import torchvision.utils

import thelper.utils

logger = logging.getLogger(__name__)


class Compose(torchvision.transforms.Compose):
    """Composes several transforms together (with support for invert ops).

    This interface is fully compatible with ``torchvision.transforms.Compose``.

    .. seealso::
        | :class:`thelper.transforms.composers.CustomStepCompose`
    """

    def __init__(self, transforms):
        """Forwards the list of transformations to the base class."""
        assert isinstance(transforms, list) and transforms, "expected transforms to be provided as a non-empty list"
        if all([isinstance(stage, dict) for stage in transforms]):
            transforms = thelper.transforms.load_transforms(transforms, avoid_transform_wrapper=True)
            transforms = transforms.transforms if isinstance(transforms, Compose) else transforms
            transforms = transforms if isinstance(transforms, list) else [transforms]
        super(Compose, self).__init__(transforms)

    def invert(self, sample):
        """Tries to invert the transformations applied to a sample.

        Will throw if one of the transformations cannot be inverted.
        """
        for t in reversed(self.transforms):
            assert hasattr(t, "invert"), f"missing invert op for transform = {repr(t)}"
            sample = t.invert(sample)
        return sample

    def __getitem__(self, idx):
        """Returns the idx-th operation wrapped by the composer."""
        assert 0 <= idx < len(self.transforms), "operation index is out of range"
        return self.transforms[idx]

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + "(transforms=[\n\t" + \
            ",\n\t".join([repr(t) for t in self.transforms]) + "\n])"

    def set_seed(self, seed):
        """Sets the internal seed to use for stochastic ops."""
        if self.transforms is not None:
            for t in self.transforms:
                if hasattr(t, "set_seed") and callable(t.set_seed):
                    t.set_seed(seed)

    def set_epoch(self, epoch=0):
        """Sets the current epoch number in order to change the behavior of some suboperations."""
        assert isinstance(epoch, int) and epoch >= 0, "invalid epoch value"
        if self.transforms is not None:
            for t in self.transforms:
                if hasattr(t, "set_epoch") and callable(t.set_epoch):
                    t.set_epoch(epoch)


class CustomStepCompose(torchvision.transforms.Compose):
    """Composes several transforms together based on an epoch schedule.

    This interface is fully compatible with ``torchvision.transforms.Compose``. It can be useful if
    some operations should change their behavior over the course of a training session. Note that
    all epoch indices are assumed to be 0-based.

    Usage example in Python::

        # We will scale the resolution of input patches based on an arbitrary schedule
        # dsize = (16, 16)   if epoch < 2         (16x16 patches before epoch 2)
        # dsize = (32, 32)   if 2 <= epoch < 4    (32x32 patches before epoch 4)
        # dsize = (64, 64)   if 4 <= epoch < 8    (64x64 patches before epoch 8)
        # dsize = (112, 112) if 8 <= epoch < 12   (112x112 patches before epoch 12)
        # dsize = (160, 160) if 12 <= epoch < 15  (160x160 patches before epoch 15)
        # dsize = (196, 196) if 15 <= epoch < 18  (196x196 patches before epoch 18)
        # dsize = (224, 224) if epoch >= 18       (224x224 patches past epoch 18)
        transforms = CustomStepCompose(milestones={
            0: thelper.transforms.Resize(dsize=(16, 16)),
            2: thelper.transforms.Resize(dsize=(32, 32)),
            4: thelper.transforms.Resize(dsize=(64, 64)),
            8: thelper.transforms.Resize(dsize=(112, 112)),
            12: thelper.transforms.Resize(dsize=(160, 160)),
            15: thelper.transforms.Resize(dsize=(196, 196)),
            18: thelper.transforms.Resize(dsize=(224, 224)),
        })
        for epoch in range(100):
            transforms.set_epoch(epoch)
            for sample in loader:
                sample = transforms(sample)
                train(...)

    Attributes:
        stages: list of epochs where a new scaling factor is to be applied.
        transforms: list of transformation to apply at each stage.
        milestones: original milestones map provided in the constructor.
        epoch: index of the current epoch.

    .. seealso::
        | :class:`thelper.transforms.composers.Compose`
    """

    def __init__(self, milestones, last_epoch=-1):
        """Receives the milestone stages (or stage lists), and the initialization state.

        If the milestones do not include the first epoch (idx = 0), then no transform will be applied
        until the next specified epoch index. When last_epoch is -1, the training is assumed to start
        from scratch.

        Args:
            milestones: Map of epoch indices tied to transformation stages. Keys must be increasing.
            last_epoch: The index of last epoch. Default: -1.
        """
        assert isinstance(milestones, dict), "milestones should be provided as a dictionary"
        self.transforms, self.stages = [], []
        if len(milestones) > 0:
            assert all([isinstance(k, int) for k in milestones.keys()]) or \
                all([isinstance(k, str) for k in milestones.keys()]), \
                "milestone stages should all be indices (integers) or strings"
            if isinstance(list(milestones.keys())[0], str):  # fixup for json-based config loaders
                self.stages = [int(key) for key in milestones.keys()]
            elif isinstance(list(milestones.keys())[0], int):
                self.stages = list(milestones.keys())
            assert self.stages == sorted(self.stages), "milestone stages should be increasing integers"
        for transforms in milestones.values():
            assert isinstance(transforms, list) or hasattr(transforms, "__call__"), \
                "stage transformations should be callable"
            if isinstance(transforms, list) and all([isinstance(t, dict) for t in transforms]):
                transforms = thelper.transforms.load_transforms(transforms, avoid_transform_wrapper=True)
            self.transforms.append(transforms)
        if 0 not in self.stages:
            self.stages.insert(0, int(0))
            self.transforms.insert(0, [])
        self.milestones = milestones  # kept only for debug/display purposes
        super(CustomStepCompose, self).__init__(self.transforms)
        self.epoch = last_epoch + 1

    def _get_stage_idx(self, epoch):
        if epoch in self.stages:
            return self.stages.index(epoch)
        return max(bisect.bisect_right(self.stages, epoch) - 1, 0)

    def invert(self, sample):
        """Tries to invert the transformations applied to a sample.

        Will throw if one of the transformations cannot be inverted.
        """
        transforms = self.transforms[self._get_stage_idx(self.epoch)]
        transforms = transforms if isinstance(transforms, list) else [transforms]
        for t in reversed(transforms):
            assert hasattr(t, "invert"), f"missing invert op for transform = {repr(t)}"
            sample = t.invert(sample)
        return sample

    def __call__(self, img):
        """Applies the current stage of transformation operations to a sample."""
        transforms = self.transforms[self._get_stage_idx(self.epoch)]
        transforms = transforms if isinstance(transforms, list) else [transforms]
        for t in transforms:
            img = t(img)
        return img

    def __getitem__(self, idx):
        """Returns the idx-th operation wrapped by the composer."""
        assert 0 <= idx < len(self.transforms), "operation index is out of range"
        return self.transforms[idx]

    def __repr__(self):
        """Provides print-friendly output for class attributes."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + "(milestones={\n\t" + \
            ",\n\t".join([f"{str(k)} :{repr(v)}" for k, v in self.milestones.items()]) + "\n})"

    def set_seed(self, seed):
        """Sets the internal seed to use for stochastic ops."""
        transforms = self.transforms[self._get_stage_idx(self.epoch)]
        transforms = transforms if isinstance(transforms, list) else [transforms]
        for t in transforms:
            if hasattr(t, "set_seed") and callable(t.set_seed):
                t.set_seed(seed)

    def set_epoch(self, epoch=0):
        """Sets the current epoch number in order to change the behavior of some suboperations."""
        assert isinstance(epoch, int) and epoch >= 0, "invalid epoch value"
        transforms = self.transforms[self._get_stage_idx(self.epoch)]
        transforms = transforms if isinstance(transforms, list) else [transforms]
        for t in transforms:
            if hasattr(t, "set_epoch") and callable(t.set_epoch):
                t.set_epoch(epoch)
        self.epoch = epoch

    def step(self, epoch=None):  # used for interface compatibility with LRSchedulers only
        """Advances the epoch tracker in order to change the behavior of some suboperations."""
        if epoch is None:
            epoch = self.epoch + 1
        self.set_epoch(epoch=epoch)



import torch

import thelper.nn


class VDSR(thelper.nn.Module):
    """Implements the VDSR architecture.

    See Kim et al., "Accurate Image Super-Resolution Using Very Deep Convolutional Networks" (2015) for more
    information (https://arxiv.org/abs/1511.04587).
    """

    def __init__(self, task, num_channels=1, base_filter=64, kernel_size0=3,
                 num_residuals=18, groups=1, activation='relu', norm='batch'):
        # note: must always forward args to base class to keep backup
        super(VDSR, self).__init__(task, **{k: v for k, v in vars().items() if k not in ["self", "task", "__class__"]})
        self.kernel_size0 = kernel_size0
        self.num_channels = num_channels
        self.input_conv = thelper.nn.common.ConvBlock(input_size=num_channels, output_size=base_filter, kernel_size=self.kernel_size0,
                                                      stride=1, padding=self.kernel_size0 // 2, norm=norm, bias=False, groups=groups,
                                                      activation=activation)
        self.num_residuals = num_residuals
        conv_blocks = []
        if self.num_residuals:
            for _ in range(self.num_residuals):
                conv_blocks.append(thelper.nn.common.ConvBlock(input_size=base_filter, output_size=base_filter, kernel_size=3,
                                                               stride=1, padding=1, norm=norm, bias=False, groups=groups,
                                                               activation=activation))
            self.residual_layers = torch.nn.Sequential(*conv_blocks)
        self.output_conv = thelper.nn.common.ConvBlock(input_size=base_filter, output_size=num_channels, kernel_size=3,
                                                       stride=1, padding=1, activation=None, norm=None, bias=False, groups=groups)
        self.weight_init()
        self.set_task(task)

    def forward(self, x):
        x0 = x.view(x.shape[0] * x.shape[1], 1, x.shape[2], x.shape[3])
        residual = x0
        x0 = self.input_conv(x0)
        if self.num_residuals > 0:
            x0 = self.residual_layers(x0)
        x0 = self.output_conv(x0)
        x0 = torch.add(x0, residual)
        x0 = x0.view(x.shape[0], x.shape[1], x.shape[2], x.shape[3])
        return x0

    def weight_init(self):
        for m in self.modules():
            thelper.nn.common.weights_init_kaiming(m)

    def set_task(self, task):
        if not isinstance(task, thelper.tasks.Regression):
            raise AssertionError("VDSR architecture only available for super res regression tasks")
        self.task = task



"""Common object interfaces module.

The interfaces defined here are fairly generic and used to eliminate
issues related to circular module importation.
"""

import copy
import logging
import os
from abc import ABC, abstractmethod
from typing import Any, AnyStr, List, Optional  # noqa: F401

import thelper.concepts
import thelper.typedefs  # noqa: F401
import thelper.utils

logger = logging.getLogger(__name__)


class PredictionConsumer(ABC):
    """Abstract model prediction consumer class.

    This interface defines basic functions required so that :class:`thelper.train.base.Trainer` can
    figure out how to instantiate and update a model prediction consumer. The most notable class derived
    from this interface is :class:`thelper.optim.metrics.Metric` which is used to monitor the
    improvement of a model during a training session. Other prediction consumers defined in
    :mod:`thelper.train.utils` will instead log predictions to local files, create graphs, etc.
    """

    def __repr__(self):
        """Returns a generic print-friendly string containing info about this consumer."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + "()"

    def reset(self):
        """Resets the internal state of the consumer.

        May be called for example by the trainer between two evaluation epochs. The default implementation
        does nothing, and if a reset behavior is needed, it should be implemented by the derived class.
        """
        pass

    @abstractmethod
    def update(self,        # see `thelper.typedefs.IterCallbackParams` for more info
               task,        # type: thelper.tasks.utils.Task
               input,       # type: thelper.typedefs.InputType
               pred,        # type: thelper.typedefs.AnyPredictionType
               target,      # type: thelper.typedefs.AnyTargetType
               sample,      # type: thelper.typedefs.SampleType
               loss,        # type: Optional[float]
               iter_idx,    # type: int
               max_iters,   # type: int
               epoch_idx,   # type: int
               max_epochs,  # type: int
               **kwargs,    # type: Any
               ):           # type: (...) -> None
        """Receives the latest prediction and groundtruth tensors from the training session.

        The data given here will be "consumed" internally, but it should NOT be modified. For example,
        a classification accuracy metric would accumulate the correct number of predictions in comparison
        to groundtruth labels, while a plotting logger would add new corresponding dots to a curve.

        Remember that input, prediction, and target tensors received here will all have a batch dimension!

        The exact signature of this function should match the one of the callbacks defined in
        :class:`thelper.train.base.Trainer` and specified by ``thelper.typedefs.IterCallbackParams``.
        """
        raise NotImplementedError


class ClassNamesHandler(ABC):
    """Generic interface to handle class names operations for inheriting classes.

    Attributes:
        class_names: holds the list of class label names.
        class_indices: holds a mapping (dict) of class-names-to-label-indices.
    """

    # args and kwargs are for additional inputs that could be passed down involuntarily, but that are not necessary
    def __init__(self, class_names=None, *args, **kwargs):
        # type: (Optional[List[AnyStr]], Any, Any) -> None
        """Initializes the class names array, if an object is provided."""
        self.class_names = class_names

    @property
    def class_names(self):
        """Returns the list of class names considered "of interest" by the derived class."""
        return self._class_names

    @class_names.setter
    def class_names(self, class_names):
        """Sets the list of class names considered "of interest" by the derived class."""
        if class_names is None:
            self._class_names = None
            self._class_indices = None
            return
        if isinstance(class_names, str) and os.path.exists(class_names):
            class_names = thelper.utils.load_config(class_names, add_name_if_missing=False)
        if isinstance(class_names, dict):
            indices_as_keys = [idx in class_names or str(idx) in class_names
                               for idx in range(len(class_names))]
            indices_as_values = [idx in class_names.values() or str(idx) in class_names.values()
                                 for idx in range(len(class_names))]
            missing_indices = {idx: not (a or b) for idx, a, b in
                               zip(range(len(class_names)), indices_as_keys, indices_as_values)}
            assert not any(missing_indices.values()), \
                f"labeling is not contiguous, missing indices:\n\t{[k for k, v in missing_indices.items() if v]}"
            assert all(indices_as_keys) or all(indices_as_values), "cannot mix indices in keys and values"
            if all(indices_as_keys):
                class_names = [thelper.utils.get_key([idx, str(idx)], class_names)
                               for idx in range(len(class_names))]
            elif all(indices_as_values):
                class_names = [k for idx in range(len(class_names))
                               for k, v in class_names.items() if v == idx or v == str(idx)]
        assert isinstance(class_names, list), "expected class names to be provided as an array"
        assert all([isinstance(name, str) for name in class_names]), "all classes must be named with strings"
        assert len(class_names) >= 1, "should have at least one class!"
        if len(class_names) != len(set(class_names)):
            # no longer throwing here, imagenet possesses such a case ('crane#134' and 'crane#517')
            logger.warning("found duplicated name in class list, might be a data entry problem...")
            class_names = [name if class_names.count(name) == 1 else name + "#" + str(idx)
                           for idx, name in enumerate(class_names)]
        self._class_names = copy.deepcopy(class_names)
        self._class_indices = {class_name: idx for idx, class_name in enumerate(class_names)}

    @property
    def class_indices(self):
        """Returns the class-name-to-index map used for encoding labels as integers."""
        return self._class_indices

    @class_indices.setter
    def class_indices(self, class_indices):
        """Sets the class-name-to-index map used for encoding labels as integers."""
        assert class_indices is None or isinstance(class_indices, dict), "indices must be provided as dictionary"
        self.class_names = class_indices


class FormatHandler(ABC):
    """Generic interface to handle format output operations for inheriting classes.

    If :attr:`format` is specified and matches a supported one (with a matching ``report_<format>`` method), this
    method is used to generate the output. Defaults to ``"text"`` if not specified or provided value is not found
    within supported formatting methods.

    Attributes:
        format: format to be used for producing the report (default: "text")
        ext: extension associated with generated format (default: "txt")
    """

    # corresponding formats should have preferred extension last
    # extension -> format
    __formats__ = {
        "text": "text",
        "txt": "text",
        "csv": "csv",
        "yaml": "yaml",
        "yml": "yaml",
        "json": "json",
        "geojson": "geojson",
    }
    # format -> extension
    __fmt_ext__ = {fmt: ext for ext, fmt in __formats__.items()}

    # args and kwargs are for additional inputs that could be passed down involuntarily, but that are not necessary
    def __init__(self, format="text", *args, **kwargs):
        # type: (AnyStr, Any, Any) -> None
        self.format = None
        self.ext = None
        self.solve_format(format)

    def solve_format(self, format):
        # type: (Optional[AnyStr]) -> None
        self.format = self.__formats__.get(format, "text")
        self.ext = self.__fmt_ext__.get(self.format, "txt")

    def report(self, format=None):
        # type: (AnyStr) -> Optional[AnyStr]
        """
        Returns the report as a print-friendly string, matching the specified format if specified in configuration.

        Args:
            format: format to be used for producing the report (default: initialization attribute or "text" if invalid)
        """
        self.solve_format(format or self.format or "text")
        if isinstance(self.format, str):
            formatter = getattr(self, f"report_{self.format.lower()}", None)
            if formatter is not None:
                return formatter()
        return self.report_text()

    @abstractmethod
    def report_text(self):
        # type: () -> Optional[AnyStr]
        """Must be implemented by inheriting classes. Default report text representation."""
        raise NotImplementedError



"""
Command-line module, for use with a ``__main__`` entrypoint.

This module contains the primary functions used to create or resume a training session, to start a
visualization session, or to start an annotation session. The basic argument that needs to be provided
by the user to create any kind of session is a configuration dictionary. For sessions that produce
outputs, the path to a directory where to save the data is also needed.
"""

import argparse
import logging
import os
from typing import Any, Union

import torch
import tqdm

import thelper

TASK_COMPAT_CHOICES = frozenset(["old", "new", "compat"])


def create_session(config, save_dir):
    """Creates a session to train a model.

    All generated outputs (model checkpoints and logs) will be saved in a directory named after the
    session (the name itself is specified in ``config``), and located in ``save_dir``.

    Args:
        config: a dictionary that provides all required data configuration and trainer parameters; see
            :class:`thelper.train.base.Trainer` and :func:`thelper.data.utils.create_loaders` for more information.
            Here, it is only expected to contain a ``name`` field that specifies the name of the session.
        save_dir: the path to the root directory where the session directory should be saved. Note that
            this is not the path to the session directory itself, but its parent, which may also contain
            other session directories.

    .. seealso::
        | :class:`thelper.train.base.Trainer`
    """
    logger = thelper.utils.get_func_logger()
    session_name = thelper.utils.get_config_session_name(config)
    assert session_name is not None, "config missing 'name' field required for output directory"
    logger.info("creating new training session '%s'..." % session_name)
    thelper.utils.setup_globals(config)
    save_dir = thelper.utils.get_save_dir(save_dir, session_name, config)
    logger.debug("session will be saved at '%s'" % os.path.abspath(save_dir).replace("\\", "/"))
    task, train_loader, valid_loader, test_loader = thelper.data.create_loaders(config, save_dir)
    model = thelper.nn.create_model(config, task, save_dir=save_dir)
    loaders = (train_loader, valid_loader, test_loader)
    trainer = thelper.train.create_trainer(session_name, save_dir, config, model, task, loaders)
    logger.debug("starting trainer")
    if train_loader:
        trainer.train()
    else:
        trainer.eval()
    logger.debug("all done")
    return trainer.outputs


def resume_session(ckptdata, save_dir, config=None, eval_only=False, task_compat=None):
    """Resumes a previously created training session.

    Since the saved checkpoints contain the original session's configuration, the ``config`` argument
    can be set to ``None`` if the session should simply pick up where it was interrupted. Otherwise,
    the ``config`` argument can be set to a new configuration that will override the older one. This is
    useful when fine-tuning a model, or when testing on a new dataset.

    .. warning::
        If a session is resumed with an overriding configuration, the user must make sure that the
        inputs/outputs of the older model are compatible with the new parameters. For example, with
        classifiers, this means that the number of classes parsed by the dataset (and thus to be
        predicted by the model) should remain the same. This is a limitation of the framework that
        should be addressed in a future update.

    .. warning::
        A resumed session will not be compatible with its original RNG states if the number of workers
        used is changed. To get 100% reproducible results, make sure you run with the same worker count.

    Args:
        ckptdata: raw checkpoint data loaded via ``torch.load()``; it will be parsed by the various
            parts of the framework that need to reload their previous state.
        save_dir: the path to the root directory where the session directory should be saved. Note that
            this is not the path to the session directory itself, but its parent, which may also contain
            other session directories.
        config: a dictionary that provides all required data configuration and trainer parameters; see
            :class:`thelper.train.base.Trainer` and :func:`thelper.data.utils.create_loaders` for more information.
            Here, it is only expected to contain a ``name`` field that specifies the name of the session.
        eval_only: specifies whether training should be resumed or the model should only be evaluated.
        task_compat: specifies how to handle discrepancy between old task from checkpoint and new task from config

    .. seealso::
        | :class:`thelper.train.base.Trainer`
    """
    logger = thelper.utils.get_func_logger()
    if ckptdata is None or not ckptdata:
        raise AssertionError("must provide valid checkpoint data to resume a session!")
    if not config:
        if "config" not in ckptdata or not ckptdata["config"]:
            raise AssertionError("checkpoint data missing 'config' field")
        config = ckptdata["config"]
    session_name = thelper.utils.get_config_session_name(config)
    assert session_name is not None, "config missing 'name' field required for output directory"
    logger.info("loading training session '%s' objects..." % session_name)
    thelper.utils.setup_globals(config)
    save_dir = thelper.utils.get_save_dir(save_dir, session_name, config, resume=True)
    logger.debug("session will be saved at '%s'" % os.path.abspath(save_dir).replace("\\", "/"))
    new_task, train_loader, valid_loader, test_loader = thelper.data.create_loaders(config, save_dir)
    if "task" not in ckptdata or not ckptdata["task"] or not isinstance(ckptdata["task"], (thelper.tasks.Task, str)):
        raise AssertionError("invalid checkpoint, cannot reload previous model task")
    old_task = thelper.tasks.create_task(ckptdata["task"]) if isinstance(ckptdata["task"], str) else ckptdata["task"]
    if not old_task.check_compat(new_task, exact=True):
        compat_task = None if not old_task.check_compat(new_task) else old_task.get_compat(new_task)
        if task_compat in TASK_COMPAT_CHOICES:
            logger.warning("discrepancy between old task from checkpoint and new task from config resolved by " +
                           f"input argument: task_compat={task_compat}")
            task_compat_mode = task_compat
        else:
            loaders_config = thelper.utils.get_key(["data_config", "loaders"], config)
            task_compat_mode = thelper.utils.get_key_def("task_compat_mode", loaders_config, default=None)
            if task_compat_mode in TASK_COMPAT_CHOICES:
                logger.warning("discrepancy between old task from checkpoint and new task from config resolved by " +
                               f"config: task_compat_mode={task_compat_mode}")
        if task_compat_mode not in TASK_COMPAT_CHOICES:
            task_compat_mode = thelper.utils.query_string(
                "Found discrepancy between old task from checkpoint and new task from config; " +
                "which one would you like to resume the session with?\n" +
                f"\told: {str(old_task)}\n\tnew: {str(new_task)}\n" +
                (f"\tcompat: {str(compat_task)}\n\n" if compat_task is not None else "\n") +
                "WARNING: if resuming with new or compat, some weights might be discarded!",
                choices=list(TASK_COMPAT_CHOICES))
        task = old_task if task_compat_mode == "old" else new_task if task_compat_mode == "new" else compat_task
        if task_compat_mode != "old":
            # saved optimizer state might cause issues with mismatched tasks, let's get rid of it
            logger.warning("dumping optimizer state to avoid issues when resuming with modified task")
            ckptdata["optimizer"], ckptdata["scheduler"] = None, None
    else:
        task = new_task
    assert task is not None, "invalid task"
    model = thelper.nn.create_model(config, task, save_dir=save_dir, ckptdata=ckptdata)
    loaders = (None if eval_only else train_loader, valid_loader, test_loader)
    trainer = thelper.train.create_trainer(session_name, save_dir, config, model, task, loaders, ckptdata=ckptdata)
    if eval_only:
        logger.info("evaluating session '%s' checkpoint @ epoch %d" % (trainer.name, trainer.current_epoch))
        trainer.eval()
    else:
        logger.info("resuming training session '%s' @ epoch %d" % (trainer.name, trainer.current_epoch))
        trainer.train()
    logger.debug("all done")
    return trainer.outputs


def visualize_data(config):
    """Displays the images used in a training session.

    This mode does not generate any output, and is only used to visualize the (transformed) images used
    in a training session. This is useful to debug the data augmentation and base transformation pipelines
    and make sure the modified images are valid. It does not attempt to load a model or instantiate a
    trainer, meaning the related fields are not required inside ``config``.

    If the configuration dictionary includes a 'loaders' field, it will be parsed and used. Otherwise,
    if only a 'datasets' field is available, basic loaders will be instantiated to load the data. The
    'loaders' field can also be ignored if 'ignore_loaders' is found within the 'viz' section of the config
    and set to ``True``. Each minibatch will be displayed via pyplot or OpenCV. The display will block and
    wait for user input, unless 'block' is set within the 'viz' section's 'kwargs' config as ``False``.

    Args:
        config: a dictionary that provides all required data configuration parameters; see
            :func:`thelper.data.utils.create_loaders` for more information.

    .. seealso::
        | :func:`thelper.data.utils.create_loaders`
        | :func:`thelper.data.utils.create_parsers`
    """
    logger = thelper.utils.get_func_logger()
    logger.info("creating visualization session...")
    thelper.utils.setup_globals(config)
    viz_config = thelper.utils.get_key_def("viz", config, default={})
    if not isinstance(viz_config, dict):
        raise AssertionError("unexpected viz config type")
    ignore_loaders = thelper.utils.get_key_def("ignore_loaders", viz_config, default=False)
    viz_kwargs = thelper.utils.get_key_def(["params", "kwargs"], viz_config, default={})
    if not isinstance(viz_kwargs, dict):
        raise AssertionError("unexpected viz kwargs type")
    if thelper.utils.get_key_def(["data_config", "loaders"], config, default=None) is None or ignore_loaders:
        datasets, task = thelper.data.create_parsers(config)
        loader_map = {dataset_name: thelper.data.DataLoader(dataset,) for dataset_name, dataset in datasets.items()}
        # we assume no transforms were done in the parser, and images are given as read by opencv
        viz_kwargs["ch_transpose"] = thelper.utils.get_key_def("ch_transpose", viz_kwargs, False)
        viz_kwargs["flip_bgr"] = thelper.utils.get_key_def("flip_bgr", viz_kwargs, False)
    else:
        task, train_loader, valid_loader, test_loader = thelper.data.create_loaders(config)
        loader_map = {"train": train_loader, "valid": valid_loader, "test": test_loader}
        # we assume transforms were done in the loader, and images are given as expected by pytorch
        viz_kwargs["ch_transpose"] = thelper.utils.get_key_def("ch_transpose", viz_kwargs, True)
        viz_kwargs["flip_bgr"] = thelper.utils.get_key_def("flip_bgr", viz_kwargs, False)
    redraw = None
    viz_kwargs["block"] = thelper.utils.get_key_def("block", viz_kwargs, default=True)
    assert "quit" not in loader_map
    choices = list(loader_map.keys()) + ["quit"]
    while True:
        choice = thelper.utils.query_string("Which loader would you like to visualize?", choices=choices)
        if choice == "quit":
            break
        loader = loader_map[choice]
        if loader is None:
            logger.info("loader '%s' is empty" % choice)
            continue
        batch_count = len(loader)
        logger.info("initializing loader '%s' with %d batches..." % (choice, batch_count))
        for sample in tqdm.tqdm(loader):
            input = sample[task.input_key]
            target = sample[task.gt_key] if task.gt_key in sample else None
            redraw = thelper.draw.draw(task=task, input=input, target=target, redraw=redraw, **viz_kwargs)
        logger.info("all done")


def annotate_data(config, save_dir):
    """Launches an annotation session for a dataset using a specialized GUI tool.

    Note that the annotation type must be supported by the GUI tool. The annotations created by the user
    during the session will be saved in the session directory.

    Args:
        config: a dictionary that provides all required dataset and GUI tool configuration parameters; see
            :func:`thelper.data.utils.create_parsers` and :func:`thelper.gui.utils.create_annotator` for more
            information.
        save_dir: the path to the root directory where the session directory should be saved. Note that
            this is not the path to the session directory itself, but its parent, which may also contain
            other session directories.

    .. seealso::
        | :func:`thelper.gui.annotators.Annotator`
        | :func:`thelper.gui.annotators.ImageSegmentAnnotator`
    """
    # import gui here since it imports packages that will cause errors in CLI-only environments
    import thelper.gui
    logger = thelper.utils.get_func_logger()
    session_name = thelper.utils.get_config_session_name(config)
    assert session_name is not None, "config missing 'name' field required for output directory"
    logger.info("creating annotation session '%s'..." % session_name)
    thelper.utils.setup_globals(config)
    save_dir = thelper.utils.get_save_dir(save_dir, session_name, config)
    logger.debug("session will be saved at '%s'" % os.path.abspath(save_dir).replace("\\", "/"))
    datasets, _ = thelper.data.create_parsers(config)
    annotator = thelper.gui.create_annotator(session_name, save_dir, config, datasets)
    logger.debug("starting annotator")
    annotator.run()
    logger.debug("all done")


def split_data(config, save_dir):
    """Launches a dataset splitting session.

    This mode will generate an HDF5 archive that contains the split datasets defined in the session
    configuration file. This archive can then be reused in a new training session to guarantee a fixed
    distribution of training, validation, and testing samples. It can also be used outside the framework
    in order to reproduce an experiment.

    The configuration dictionary must minimally contain two sections: 'datasets' and 'loaders'. A third
    section, 'split', can be used to provide settings regarding the archive packing and compression
    approaches to use.

    The HDF5 archive will be saved in the session's output directory.

    Args:
        config: a dictionary that provides all required data configuration parameters; see
            :func:`thelper.data.utils.create_loaders` for more information.
        save_dir: the path to the root directory where the session directory should be saved. Note that
            this is not the path to the session directory itself, but its parent, which may also contain
            other session directories.

    .. seealso::
        | :func:`thelper.data.utils.create_loaders`
        | :func:`thelper.data.utils.create_hdf5`
        | :class:`thelper.data.parsers.HDF5Dataset`
    """
    logger = thelper.utils.get_func_logger()
    session_name = thelper.utils.get_config_session_name(config)
    assert session_name is not None, "config missing 'name' field required for output directory"
    split_config = thelper.utils.get_key_def("split", config, default={})
    if not isinstance(split_config, dict):
        raise AssertionError("unexpected split config type")
    compression = thelper.utils.get_key_def("compression", split_config, default={})
    if not isinstance(compression, dict):
        raise AssertionError("compression params should be given as dictionary")
    archive_name = thelper.utils.get_key_def("archive_name", split_config, default=(session_name + ".hdf5"))
    logger.info("creating new splitting session '%s'..." % session_name)
    thelper.utils.setup_globals(config)
    save_dir = thelper.utils.get_save_dir(save_dir, session_name, config)
    logger.debug("session will be saved at '%s'" % os.path.abspath(save_dir).replace("\\", "/"))
    task, train_loader, valid_loader, test_loader = thelper.data.create_loaders(config, save_dir)
    archive_path = os.path.join(save_dir, archive_name)
    thelper.data.create_hdf5(archive_path, task, train_loader, valid_loader, test_loader, compression, config)
    logger.debug("all done")


def export_model(config, save_dir):
    """Launches a model exportation session.

    This function will export a model defined via a configuration file into a new checkpoint that can be
    loaded elsewhere. The model can be built using the framework, or provided via its type, construction
    parameters, and weights. Its exported format will be compatible with the framework, and may also be an
    optimized/compiled version obtained using PyTorch's JIT tracer.

    The configuration dictionary must minimally contain a 'model' section that provides details on the model
    to be exported. A section named 'export' can be used to provide settings regarding the exportation
    approaches to use, and the task interface to save with the model. If a task is not explicitly defined in
    the 'export' section, the session configuration will be parsed for a 'datasets' section that can be used
    to define it. Otherwise, it must be provided through the model.

    The exported checkpoint containing the model will be saved in the session's output directory.

    Args:
        config: a dictionary that provides all required data configuration parameters; see
            :func:`thelper.nn.utils.create_model` for more information.
        save_dir: the path to the root directory where the session directory should be saved. Note that
            this is not the path to the session directory itself, but its parent, which may also contain
            other session directories.

    .. seealso::
        | :func:`thelper.nn.utils.create_model`
    """
    logger = thelper.utils.get_func_logger()
    session_name = thelper.utils.get_config_session_name(config)
    assert session_name is not None, "config missing 'name' field required for output directory"
    export_config = thelper.utils.get_key_def("export", config, default={})
    if not isinstance(export_config, dict):
        raise AssertionError("unexpected export config type")
    ckpt_name = thelper.utils.get_key_def("ckpt_name", export_config, default=(session_name + ".export.pth"))
    trace_name = thelper.utils.get_key_def("trace_name", export_config, default=(session_name + ".trace.zip"))
    save_raw = thelper.utils.get_key_def("save_raw", export_config, default=True)
    trace_input = thelper.utils.get_key_def("trace_input", export_config, default=None)
    task = thelper.utils.get_key_def("task", export_config, default=None)
    if isinstance(task, (str, dict)):
        task = thelper.tasks.create_task(task)
    if task is None and "datasets" in config:
        _, task = thelper.data.create_parsers(config)  # try to load via datasets...
    if isinstance(trace_input, str):
        trace_input = eval(trace_input)
    logger.info("exporting model '%s'..." % session_name)
    thelper.utils.setup_globals(config)
    save_dir = thelper.utils.get_save_dir(save_dir, session_name, config)
    logger.debug("exported checkpoint will be saved at '%s'" % os.path.abspath(save_dir).replace("\\", "/"))
    model = thelper.nn.create_model(config, task, save_dir=save_dir)
    if task is None:
        assert hasattr(model, "task"), "model should have task attrib if not provided already"
        task = model.task
    log_stamp = thelper.utils.get_log_stamp()
    model_type = model.get_name()
    model_params = model.config if model.config else {}
    # the saved state below should be kept compatible with the one in thelper.train.base._save
    export_state = {
        "name": session_name,
        "source": log_stamp,
        "git_sha1": thelper.utils.get_git_stamp(),
        "version": thelper.__version__,
        "task": str(task) if save_raw else task,
        "model_type": model_type,
        "model_params": model_params,
        "config": config
    }
    if trace_input is not None:
        trace_path = os.path.join(save_dir, trace_name)
        torch.jit.trace(model, trace_input).save(trace_path)
        export_state["model"] = trace_name  # will be loaded in thelper.utils.load_checkpoint
    else:
        export_state["model"] = model.state_dict() if save_raw else model
    torch.save(export_state, os.path.join(save_dir, ckpt_name))
    logger.debug("all done")


def make_argparser():
    # type: () -> argparse.ArgumentParser
    """Creates the (default) argument parser to use for the main entrypoint.

    The argument parser will contain different "operating modes" that dictate the high-level behavior of the CLI. This
    function may be modified in branches of the framework to add project-specific features.
    """
    ap = argparse.ArgumentParser(description='thelper model trainer application')
    ap.add_argument("--version", default=False, action="store_true", help="prints the version of the library and exits")
    ap.add_argument("-l", "--log", default=None, type=str, help="path to the top-level log file (default: None)")
    ap.add_argument("-v", "--verbose", action="count", default=0, help="set logging terminal verbosity level (additive)")
    ap.add_argument("--silent", action="store_true", default=False, help="deactivates all console logging activities")
    ap.add_argument("--force-stdout", action="store_true", default=False, help="force logging output to stdout instead of stderr")
    subparsers = ap.add_subparsers(title="Operating mode", dest="mode")
    new_ap = subparsers.add_parser("new", help="creates a new session from a config file")
    new_ap.add_argument("cfg_path", type=str, help="path to the session configuration file")
    new_ap.add_argument("save_dir", type=str, help="path to the session output root directory")
    cl_new_ap = subparsers.add_parser("cl_new", help="creates a new session from a config file for the cluster")
    cl_new_ap.add_argument("cfg_path", type=str, help="path to the session configuration file")
    cl_new_ap.add_argument("save_dir", type=str, help="path to the session output root directory")
    resume_ap = subparsers.add_parser("resume", help="resume a session from a checkpoint file")
    resume_ap.add_argument("ckpt_path", type=str, help="path to the checkpoint (or directory) to resume training from")
    resume_ap.add_argument("-s", "--save-dir", default=None, type=str, help="path to the session output root directory")
    resume_ap.add_argument("-m", "--map-location", default=None, help="map location for loading data (default=None)")
    resume_ap.add_argument("-c", "--override-cfg", default=None, help="override config file path (default=None)")
    resume_ap.add_argument("-e", "--eval-only", default=False, action="store_true", help="only run evaluation pass (valid+test)")
    resume_ap.add_argument("-t", "--task-compat", default=None, type=str, choices=TASK_COMPAT_CHOICES,
                           help="task compatibility mode to use to resolve any discrepancy between loaded tasks")
    viz_ap = subparsers.add_parser("viz", help="visualize the loaded data for a training/eval session")
    viz_ap.add_argument("cfg_path", type=str, help="path to the session configuration file (or session directory)")
    annot_ap = subparsers.add_parser("annot", help="launches a dataset annotation session with a GUI tool")
    annot_ap.add_argument("cfg_path", type=str, help="path to the session configuration file (or session directory)")
    annot_ap.add_argument("save_dir", type=str, help="path to the session output root directory")
    split_ap = subparsers.add_parser("split", help="launches a dataset splitting session from a config file")
    split_ap.add_argument("cfg_path", type=str, help="path to the session configuration file (or session directory)")
    split_ap.add_argument("save_dir", type=str, help="path to the session output root directory")
    split_ap = subparsers.add_parser("export", help="launches a model exportation session from a config file")
    split_ap.add_argument("cfg_path", type=str, help="path to the session configuration file (or session directory)")
    split_ap.add_argument("save_dir", type=str, help="path to the session output root directory")
    return ap


def setup(args=None, argparser=None):
    # type: (Any, argparse.ArgumentParser) -> Union[int, argparse.Namespace]
    """Sets up the argument parser (if not already done externally) and parses the input CLI arguments.

    This function may return an error code (integer) if the program should exit immediately. Otherwise, it will return
    the parsed arguments to use in order to redirect the execution flow of the entrypoint.
    """
    argparser = argparser or make_argparser()
    args = argparser.parse_args(args=args)
    if args.version:
        print(thelper.__version__)
        return 0
    if args.mode is None:
        argparser.print_help()
        return 1
    if args.silent and args.verbose > 0:
        raise AssertionError("contradicting verbose/silent arguments provided")
    log_level = logging.INFO if args.verbose < 1 else logging.DEBUG if args.verbose < 2 else logging.NOTSET
    thelper.utils.init_logger(log_level, args.log, args.force_stdout)
    return args


def main(args=None, argparser=None):
    """Main entrypoint to use with console applications.

    This function parses command line arguments and dispatches the execution based on the selected
    operating mode. Run with ``--help`` for information on the available arguments.

    .. warning::
        If you are trying to resume a session that was previously executed using a now unavailable GPU,
        you will have to force the checkpoint data to be loaded on CPU using ``--map-location=cpu`` (or
        using ``-m=cpu``).

    .. seealso::
        | :func:`thelper.cli.create_session`
        | :func:`thelper.cli.resume_session`
        | :func:`thelper.cli.visualize_data`
        | :func:`thelper.cli.annotate_data`
        | :func:`thelper.cli.split_data`
    """
    args = setup(args=args, argparser=argparser)
    if isinstance(args, int):
        return args  # CLI must exit immediately with provided error code
    if args.mode == "new" or args.mode == "cl_new":
        thelper.logger.debug("parsing config at '%s'" % args.cfg_path)
        config = thelper.utils.load_config(args.cfg_path)
        if args.mode == "cl_new":
            trainer_config = thelper.utils.get_key_def("trainer", config, {})
            device = thelper.utils.get_key_def("device", trainer_config, None)
            if device is not None:
                raise AssertionError("cannot specify device in config for cluster sessions, it is determined at runtime")
        create_session(config, args.save_dir)
    elif args.mode == "resume":
        ckptdata = thelper.utils.load_checkpoint(args.ckpt_path, map_location=args.map_location,
                                                 always_load_latest=(not args.eval_only))
        override_config = None
        if args.override_cfg:
            thelper.logger.debug("parsing override config at '%s'" % args.override_cfg)
            override_config = thelper.utils.load_config(args.override_cfg)
        save_dir = args.save_dir
        if save_dir is None:
            save_dir = thelper.utils.get_checkpoint_session_root(args.ckpt_path)
        if save_dir is None:
            save_dir = thelper.utils.get_save_dir(out_root=None, dir_name=None, config=override_config)
        resume_session(ckptdata, save_dir, config=override_config, eval_only=args.eval_only, task_compat=args.task_compat)
    else:
        thelper.logger.debug("parsing config at '%s'" % args.cfg_path)
        config = thelper.utils.load_config(args.cfg_path)
        if args.mode == "viz":
            visualize_data(config)
        elif args.mode == "annot":
            annotate_data(config, args.save_dir)
        elif args.mode == "export":
            export_model(config, args.save_dir)
        else:  # if args.mode == "split":
            split_data(config, args.save_dir)
    return 0


if __name__ == "__main__":
    main()



"""Classification task interface module.

This module contains a class that defines the objectives of models/trainers for classification tasks.
"""
import logging
from typing import Optional  # noqa: F401

import numpy as np
import torch

import thelper.concepts
import thelper.utils
from thelper.ifaces import ClassNamesHandler
from thelper.tasks.utils import Task

logger = logging.getLogger(__name__)


@thelper.concepts.classification
class Classification(Task, ClassNamesHandler):
    """Interface for input labeling/classification tasks.

    This specialization requests that when given an input tensor, the trained model should
    provide prediction scores for each predefined label (or class). The label names are used
    here to help categorize samples, and to assure that two tasks are only identical when their
    label counts and ordering match.

    Attributes:
        class_names: list of label (class) names to predict (each name should be a string).
        input_key: the key used to fetch input tensors from a sample dictionary.
        label_key: the key used to fetch label (class) names/indices from a sample dictionary.
        meta_keys: the list of extra keys provided by the data parser inside each sample.

    .. seealso::
        | :class:`thelper.tasks.utils.Task`
        | :class:`thelper.train.classif.ImageClassifTrainer`
    """

    def __init__(self, class_names, input_key, label_key, meta_keys=None):
        """Receives and stores the class (or label) names to predict, the input tensor key, the
        groundtruth label (class) key, and the extra (meta) keys produced by the dataset parser(s).

        The class names can be provided as a list of strings, or as a path to a json file that
        contains such a list. The list must contain at least two items. All other arguments are
        used as-is to index dictionaries, and must therefore be key-compatible types.
        """
        super(Classification, self).__init__(input_key, label_key, meta_keys)
        ClassNamesHandler.__init__(self, class_names=class_names)

    def get_class_sizes(self, samples):
        """Given a list of samples, returns a map of sample counts for each class label."""
        class_idxs = self.get_class_sample_map(samples)
        return {class_name: len(class_idxs[class_name]) for class_name in class_idxs}

    def get_class_sample_map(self, samples, unset_key=None):
        """Splits a list of samples based on their labels into a map of sample lists.

        This function is useful if we need to split a dataset based on its label categories in
        order to sort it, augment it, or re-balance it. The samples do not need to be fully loaded
        for this to work, as only their label (gt) value will be queried. If a sample is missing
        its label, it will be ignored and left out of the generated dictionary unless a value is
        given for ``unset_key``.

        Args:
            samples: the samples to split, where each sample is provided as a dictionary.
            unset_key: a key under which all unlabeled samples should be kept (``None`` = ignore).

        Returns:
            A dictionary that maps each class label to its corresponding list of samples.
        """
        assert samples is not None and isinstance(samples, list), "invalid sample list "
        sample_idxs = {class_name: [] for class_name in self.class_names}
        import collections
        if unset_key is not None:
            assert isinstance(unset_key, collections.abc.Hashable), "unset class name key should be hashable"
            assert unset_key not in sample_idxs, "unset class name key cannot already be in class names list"
            sample_idxs[unset_key] = []
        for sample_idx, sample in enumerate(samples):
            if self.gt_key is None or self.gt_key not in sample:
                if unset_key is not None:
                    class_name = unset_key
                else:
                    continue
            else:
                class_name = sample[self.gt_key]
                assert isinstance(class_name, (str, int, np.ndarray, torch.Tensor)) and thelper.utils.is_scalar(class_name), \
                    "unexpected sample label type (need scalar, string or int)"
                if isinstance(class_name, str):
                    assert class_name in self.class_names, f"label '{class_name}' not found in original task class names list"
                else:
                    if isinstance(class_name, torch.Tensor):
                        class_name = class_name.item()
                    # dataset must already be using indices, we will forgive this...
                    # (this is pretty much always the case for torchvision datasets)
                    assert 0 <= class_name < len(self.class_names), "class name given as out-of-range index"
                    class_name = self.class_names[class_name]
            sample_idxs[class_name].append(sample_idx)
        return sample_idxs

    def check_compat(self, task, exact=False):
        # type: (Classification, Optional[bool]) -> bool
        """Returns whether the current task is compatible with the provided one or not.

        This is useful for sanity-checking, and to see if the inputs/outputs of two models
        are compatible. If ``exact = True``, all fields will be checked for exact (perfect)
        compatibility (in this case, matching meta keys and class name order).
        """
        if isinstance(task, Classification):
            # if both tasks are related to classification, gt keys and class names must match
            return self.input_key == task.input_key and \
                (self.gt_key is None or task.gt_key is None or self.gt_key == task.gt_key) and \
                all([cls in self.class_names for cls in task.class_names]) and \
                (not exact or (self.class_names == task.class_names and
                               set(self.meta_keys) == set(task.meta_keys) and
                               self.gt_key == task.gt_key))
        elif type(task) == Task:
            # if 'task' simply has no gt, compatibility rests on input key only
            return not exact and self.input_key == task.input_key and task.gt_key is None
        return False

    def get_compat(self, task):
        """Returns a task instance compatible with the current task and the given one."""
        assert isinstance(task, Classification) or type(task) == Task, \
            f"cannot create compatible task from types '{type(task)}' and '{type(self)}'"
        if isinstance(task, Classification):
            assert self.input_key == task.input_key, "input key mismatch, cannot create compatible task"
            assert self.gt_key is None or task.gt_key is None or self.gt_key == task.gt_key, \
                "gt key mismatch, cannot create compatible task"
            meta_keys = list(set(self.meta_keys + task.meta_keys))
            # cannot use set for class names, order needs to stay intact!
            class_names = self.class_names + [name for name in task.class_names if name not in self.class_names]
            return Classification(class_names=class_names, input_key=self.input_key,
                                  label_key=self.gt_key, meta_keys=meta_keys)
        elif type(task) == Task:
            assert self.check_compat(task), f"cannot create compatible task between:\n\t{str(self)}\n\t{str(task)}"
            meta_keys = list(set(self.meta_keys + task.meta_keys))
            return Classification(class_names=self.class_names, input_key=self.input_key,
                                  label_key=self.gt_key, meta_keys=meta_keys)

    def __repr__(self):
        """Creates a print-friendly representation of a classification task."""
        return self.__class__.__module__ + "." + self.__class__.__qualname__ + \
            f"(class_names={repr(self.class_names)}, input_key={repr(self.input_key)}, " + \
            f"label_key={repr(self.gt_key)}, meta_keys={repr(self.meta_keys)})"



import os
import shutil

import pytest

test_save_path = ".pytest_cache"

dummy_save_path = os.path.join(test_save_path, "dummy")
dummy_config_path = os.path.join(dummy_save_path, "dummy.json")
dummy_ckpt_path = os.path.join(dummy_save_path, "checkpoints", "dummy.ckpt")


def test_main_init(mocker):
    import thelper.__main__
    import thelper.cli
    _ = mocker.patch.object(thelper.cli, "main", return_value=1337)
    _ = mocker.patch.object(thelper.__main__, "__name__", "__main__")
    mock_exit = mocker.patch.object(thelper.__main__.sys, "exit")
    thelper.__main__.init()
    assert mock_exit.call_args[0][0] == 1337


@pytest.fixture
def dummy_config(request):
    def fin():
        shutil.rmtree(dummy_save_path, ignore_errors=True)
    fin()
    request.addfinalizer(fin)
    os.makedirs(os.path.join(dummy_save_path, "logs"), exist_ok=True)
    os.makedirs(os.path.join(dummy_save_path, "checkpoints"), exist_ok=True)
    os.makedirs(os.path.join(dummy_save_path, "bad/dir/no/ckpts"), exist_ok=True)
    open(dummy_config_path, "a").close()
    open(dummy_ckpt_path, "a").close()
    return


def test_main_args(dummy_config, mocker):
    mock_create = mocker.patch("thelper.cli.create_session")
    _ = mocker.patch("thelper.cli.resume_session")
    _ = mocker.patch("thelper.cli.visualize_data")
    _ = mocker.patch("thelper.cli.annotate_data")
    _ = mocker.patch("thelper.cli.split_data")
    _ = mocker.patch("thelper.cli.export_model")
    from thelper.cli import main
    assert main([]) == 1
    assert main(["--version"]) == 0
    with pytest.raises(AssertionError):
        main(["-v", "--silent", "new", dummy_config_path, dummy_save_path])
    config = {}

    def config_getter(*args, **kwargs):
        nonlocal config
        return config

    mock_config_load = mocker.patch("thelper.utils.load_config")
    mock_config_load.side_effect = config_getter
    config = {"oii": "test"}
    assert main(["new", dummy_config_path, dummy_save_path]) == 0
    assert mock_create.called_with(config, dummy_save_path)
    assert main(["cl_new", dummy_config_path, dummy_save_path]) == 0
    config = {"trainer": {"device": "cpu"}}
    with pytest.raises(AssertionError):
        _ = main(["cl_new", dummy_config_path, dummy_save_path])
    assert main(["viz", dummy_config_path]) == 0
    assert main(["annot", dummy_config_path, dummy_save_path]) == 0
    assert main(["split", dummy_config_path, dummy_save_path]) == 0
    assert main(["export", dummy_config_path, dummy_save_path]) == 0
    _ = mocker.patch("thelper.utils.load_checkpoint")
    assert main(["resume", dummy_save_path]) == 0
    assert main(["resume", dummy_save_path, "-c=" + dummy_config_path]) == 0
    assert main(["resume", os.path.join(dummy_save_path, "checkpoints"), "-c=" + dummy_config_path]) == 0
    mocker_getter = mocker.patch("thelper.utils.get_save_dir")
    assert main(["resume", dummy_save_path, "-c=" + dummy_config_path]) == 0
    assert mocker_getter.call_count == 0
    assert main(["resume", os.path.join(dummy_save_path, "bad/dir/no/ckpts")]) == 0
    assert mocker_getter.call_count == 1
    assert main(["resume", dummy_ckpt_path]) == 0
    assert main(["resume", dummy_ckpt_path, "-s=" + dummy_save_path]) == 0



# MobileNet v2 derived from https://github.com/tonylins/pytorch-mobilenet-v2
import math

import torch.nn as nn

import thelper


def conv_bn(inp, oup, stride):
    return nn.Sequential(
        nn.Conv2d(inp, oup, 3, stride, 1, bias=False),
        nn.BatchNorm2d(oup),
        nn.ReLU6(inplace=True)
    )


def conv_1x1_bn(inp, oup):
    return nn.Sequential(
        nn.Conv2d(inp, oup, 1, 1, 0, bias=False),
        nn.BatchNorm2d(oup),
        nn.ReLU6(inplace=True)
    )


class InvertedResidual(nn.Module):

    def __init__(self, inp, oup, stride, expand_ratio):
        super().__init__()
        self.stride = stride
        if stride not in [1, 2]:
            raise AssertionError("stride should be 1 or 2")
        self.use_res_connect = self.stride == 1 and inp == oup
        self.conv = nn.Sequential(
            # pw
            nn.Conv2d(inp, inp * expand_ratio, 1, 1, 0, bias=False),
            nn.BatchNorm2d(inp * expand_ratio),
            nn.ReLU6(inplace=True),
            # dw
            nn.Conv2d(inp * expand_ratio, inp * expand_ratio, 3, stride, 1, groups=inp * expand_ratio, bias=False),
            nn.BatchNorm2d(inp * expand_ratio),
            nn.ReLU6(inplace=True),
            # pw-linear
            nn.Conv2d(inp * expand_ratio, oup, 1, 1, 0, bias=False),
            nn.BatchNorm2d(oup),
        )

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


class MobileNetV2(thelper.nn.Module):

    def __init__(self, task, input_size=224, width_mult=1.):
        # note: must always forward args to base class to keep backup
        super().__init__(task, input_size=input_size, width_mult=width_mult)
        # setting of inverted residual blocks
        self.interverted_residual_setting = [
            # t, c, n, s
            [1, 16, 1, 1],
            [6, 24, 2, 2],
            [6, 32, 3, 2],
            [6, 64, 4, 2],
            [6, 96, 3, 1],
            [6, 160, 3, 2],
            [6, 320, 1, 1],
        ]
        # building first layer
        if input_size % 32 != 0:
            raise AssertionError("input size should be multiple of 32 px")
        input_channel = int(32 * width_mult)
        self.last_channel = int(1280 * width_mult) if width_mult > 1.0 else 1280
        self.features = [conv_bn(3, input_channel, 2)]
        # building inverted residual blocks
        for t, c, n, s in self.interverted_residual_setting:
            output_channel = int(c * width_mult)
            for i in range(n):
                if i == 0:
                    self.features.append(InvertedResidual(input_channel, output_channel, s, t))
                else:
                    self.features.append(InvertedResidual(input_channel, output_channel, 1, t))
                input_channel = output_channel
        # building last several layers
        self.features.append(conv_1x1_bn(input_channel, self.last_channel))
        self.features.append(nn.AvgPool2d(input_size // 32))
        # make it nn.Sequential
        self.features = nn.Sequential(*self.features)
        # building classifier
        self.classif_features = 1000  # default count, will be updated if needed
        self.classifier = nn.Sequential(
            nn.Dropout(),
            nn.Linear(self.last_channel, self.classif_features),
        )
        self._initialize_weights()
        self.set_task(task)

    def forward(self, x):
        x = self.features(x)
        x = x.view(-1, self.last_channel)
        x = self.classifier(x)
        return x

    def set_task(self, task):
        assert isinstance(task, thelper.tasks.Classification), "missing impl for non-classif task type"
        num_classes = len(task.class_names)
        if num_classes != self.classif_features:
            self.classifier = nn.Sequential(
                nn.Dropout(),
                nn.Linear(self.last_channel, num_classes),
            )
            self.classif_features = num_classes
        self.task = task

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                n = m.weight.size(1)
                m.weight.data.normal_(0, 0.01)
                m.bias.data.zero_()



"""Geospatial dataset parsing/loading package.

This package contains classes and functions whose role is to fetch the data required to train, validate,
and test a model on geospatial data. Importing the modules inside this package requires GDAL.
"""

import logging

import thelper.data.geo.parsers  # noqa: F401
import thelper.data.geo.utils  # noqa: F401

logger = logging.getLogger("thelper.data.geo")



"""Evaluation classes/funcs module.

This module contains procedures used to evaluate models and prediction results on specific
tasks or datasets. These procedures may be used as part of metric classes (defined in
:mod:`thelper.optim.metrics`) or high-level debug/drawing utilities.
"""

import numpy as np

import thelper


def compute_iou(bbox1, bbox2):
    # type: (thelper.tasks.detect.BoundingBox, thelper.tasks.detect.BoundingBox) -> float
    """Computes and returns the Intersection over Union (IoU) of two bounding boxes."""
    assert isinstance(bbox1, thelper.data.BoundingBox) and isinstance(bbox2, thelper.data.BoundingBox), \
        "unexpected input bounding box types"
    bbox1_marg = 1 if bbox1.include_margin else 0
    bbox2_marg = 1 if bbox2.include_margin else 0
    intersection_width = min(bbox1.right + bbox1_marg, bbox2.right + bbox2_marg) - max(bbox1.left, bbox2.left)
    intersection_height = min(bbox1.bottom + bbox1_marg, bbox2.bottom + bbox2_marg) - max(bbox1.top, bbox2.top)
    intersection_area = max(0, intersection_width) * max(0, intersection_height)
    return float(intersection_area / float(bbox1.area + bbox2.area - intersection_area))


def compute_pascalvoc_metrics(pred_bboxes, gt_bboxes, task, iou_threshold=0.5, method="all-points"):
    """Computes the metrics used by the VOC Pascal 2012 challenge.

    This function is inspired from the 'Object Detection Metrics' repository of Rafael Padilla.
    See https://github.com/rafaelpadilla/Object-Detection-Metrics for more information.
    The original code is distributed under the MIT License, Copyright (c) 2018 Rafael Padilla.

    Args:
        pred_bboxes: list of bbox predictions generated by the model under evaluation.
        gt_bboxes: list of groundtruth bounding boxes defined by the dataset.
        task: task definition object that holds a vector of all class names.
        iou_threshold: Intersection Over Union (IOU) threshold for true/false positive classification.
        method: the evaluation method to use; can be the the latest & official PASCAL VOC toolkit
            approach ("all-points"), or the 11-point approach ("11-points") described in the original
            paper ("The PASCAL Visual Object Classes(VOC) Challenge").

    Returns:
        A dictionary containing evaluation information and metrics for each class. Each entry contains:
        - ``precision``: array with the precision values;
        - ``recall``: array with the recall values;
        - ``AP``: average precision;
        - ``interpolated precision``: interpolated precision values;
        - ``interpolated recall``: interpolated recall values;
        - ``total positives``: total number of ground truth positives;
        - ``total TP``: total number of True Positive detections;
        - ``total FP``: total number of False Negative detections.
    """
    assert isinstance(pred_bboxes, (list, np.ndarray)) and all([isinstance(b, thelper.data.BoundingBox) for b in pred_bboxes]), \
        "invalid predictions format (expected list of bounding box objects)"
    assert all([isinstance(bbox.confidence, float) and 0 <= bbox.confidence <= 1 for bbox in pred_bboxes]), \
        "predicted bounding boxes must be provided with confidence values in [0,1]"
    assert all([bbox.image_id is not None for bbox in pred_bboxes]), "predicted bbox image id must be defined"
    assert isinstance(gt_bboxes, (list, np.ndarray)) and all([isinstance(b, thelper.data.BoundingBox) for b in gt_bboxes]), \
        "invalid input groundtruth format (expected list of bounding box objects)"
    assert all([bbox.image_id is not None for bbox in gt_bboxes]), "gt bbox image id must be defined"
    assert isinstance(task, thelper.tasks.Detection) and task.class_names, "invalid task object (should be detection)"
    assert 0 < iou_threshold <= 1, "invalid intersection over union value (should be in ]0,1])"
    assert method in ["all-points", "11-points"], "invalid method (should be 'all-points' or '11-points')"
    image_ids = list(set([bbox.image_id for bbox in pred_bboxes]) | set([bbox.image_id for bbox in gt_bboxes]))
    image_ids = {k: idx for idx, k in enumerate(image_ids)}
    gt_used_flags = [[[[bbox, False] for bbox in gt_bboxes
                       if (((isinstance(bbox.class_id, int) and bbox.class_id == ci) or bbox.class_id == cn) and
                           bbox.image_id == iid)] for iid in image_ids] for ci, cn in enumerate(task.class_names)]
    ret = {}
    for class_idx, class_name in enumerate(task.class_names):
        if task.background is not None and class_name == "background":
            continue
        curr_pred_bboxes = [bbox for bbox in pred_bboxes if (isinstance(bbox.class_id, int) and bbox.class_id == class_idx) or
                            bbox.class_id == class_name]
        curr_pred_bboxes = sorted(curr_pred_bboxes, key=lambda bbox: bbox.confidence, reverse=True)
        true_positives = np.zeros(len(curr_pred_bboxes))
        false_positives = np.zeros(len(curr_pred_bboxes))
        for pred_bbox_idx, pred_bbox in enumerate(curr_pred_bboxes):
            curr_gt_bboxes = gt_used_flags[class_idx][image_ids[pred_bbox.image_id]]
            best_gt_bbox_idx, best_gt_bbox_iou = -1, float("-inf")
            for gt_bbox_idx, (gt_bbox, gt_bbox_flag) in enumerate(curr_gt_bboxes):
                iou = compute_iou(pred_bbox, gt_bbox)
                if iou > best_gt_bbox_iou:
                    best_gt_bbox_iou = iou
                    best_gt_bbox_idx = gt_bbox_idx
            if best_gt_bbox_iou >= iou_threshold:
                curr_best_gt_bbox_used_flag = curr_gt_bboxes[best_gt_bbox_idx][1]
                if not curr_best_gt_bbox_used_flag:
                    true_positives[pred_bbox_idx] = 1
                    # we can only use GT bboxes once, flag them as 'seen' after that
                    curr_gt_bboxes[best_gt_bbox_idx][1] = True
                else:
                    # if best GT bbox was already used, we discard this detection
                    # (note: we could do some combinatorial optim w/ hungarian method to solve ideally instead)
                    false_positives[pred_bbox_idx] = 1
            else:
                # if we fail to meet the minimum iou threshold, discard this detection
                false_positives[pred_bbox_idx] = 1
        true_positive_cumsum = np.cumsum(true_positives)
        npos = sum([len(bboxes) for bboxes in gt_used_flags[class_idx]])
        recall = true_positive_cumsum / npos
        precision = np.divide(true_positive_cumsum, (np.cumsum(false_positives) + true_positive_cumsum))
        avg_prec, mpre, mrec, _ = compute_average_precision(precision.tolist(), recall.tolist(), method)
        ret[class_name] = {
            "class_name": class_name,
            "iou_threshold": iou_threshold,
            "eval_method": method,
            "precision": precision,
            "recall": recall,
            "AP": avg_prec,
            "interpolated precision": mpre,
            "interpolated recall": mrec,
            "total positives": npos,
            "total TP": np.sum(true_positives),
            "total FP": np.sum(false_positives)
        }
    return ret


def compute_average_precision(precision, recall, method="all-points"):
    """Computes the average precision given an array of precision and recall values.

    This function is inspired from the 'Object Detection Metrics' repository of Rafael Padilla.
    See https://github.com/rafaelpadilla/Object-Detection-Metrics for more information.
    The original code is distributed under the MIT License, Copyright (c) 2018 Rafael Padilla.

    Args:
        precision: list of precision values for the evaluated predictions of a class.
        recall: list of recall values for the evaluated predictions of a class.
        method: the evaluation method to use; can be the the latest & official PASCAL VOC toolkit
            approach ("all-points"), or the 11-point approach ("11-points") described in the original
            paper ("The PASCAL Visual Object Classes(VOC) Challenge").

    Returns:
        A 4-element tuple containing the average precision, rectified precision/recall arrays, and
        the indices used for the integral.
    """
    assert isinstance(precision, list) and all([0 <= p <= 1 for p in precision])
    assert isinstance(recall, list) and all([0 <= r <= 1 for r in recall])
    assert method in ["all-points", "11-points"]
    if method == "all-points":
        mprecision = [0, *precision, 0]  # pad with extrema
        # run backwards through precision values, eliminate ridges
        for idx in range(len(mprecision) - 1, 0, -1):
            mprecision[idx - 1] = max(mprecision[idx - 1], mprecision[idx])
        mrecall = [0, *recall, 1]  # pad with extrema
        # eliminate duplicates
        idxs = [idx + 1 for idx in range(len(mrecall) - 1) if mrecall[1:][idx] != mrecall[0:-1][idx]]
        avg_prec = 0
        # compute integral (AUC)
        for idx in idxs:
            avg_prec = avg_prec + np.sum((mrecall[idx] - mrecall[idx - 1]) * mprecision[idx])
        return avg_prec, mprecision[0:len(mprecision) - 1], mrecall[0:len(mprecision) - 1], idxs
    else:
        mprecision = [*precision]
        rho_interp, recall_val_id = [], []
        for r in np.linspace(0, 1, 11)[::-1]:
            ridxs = np.argwhere(np.asarray(recall) >= r)
            rho_interp.append(max(mprecision[ridxs.min():]) if ridxs.size != 0 else 0)
            recall_val_id.append(r)
        avg_prec = sum(rho_interp) / 11
        rvals = [recall_val_id[0], *recall_val_id, 0]
        pvals = [0, *rho_interp, 0]
        cc = []
        for i in range(len(rvals)):
            p = (rvals[i], pvals[i - 1])
            if p not in cc:
                cc.append(p)
            p = (rvals[i], pvals[i])
            if p not in cc:
                cc.append(p)
        return [avg_prec, [i[1] for i in cc], [i[0] for i in cc], None]
